"""Various tests for the functions defined in `guppylang.prelude.quantum`."""

from typing import no_type_check
from hugr.package import ModulePointer

import guppylang.decorator
from guppylang.module import GuppyModule
from guppylang.std.angles import angle

from guppylang.std.builtins import owned, array, barrier

from guppylang.std import quantum
from guppylang.std.quantum import (
    discard,
    measure,
    qubit,
    maybe_qubit,
    measure_array,
    discard_array,
)
from guppylang.std.quantum_functional import (
    cx,
    cy,
    cz,
    h,
    t,
    s,
    x,
    y,
    z,
    tdg,
    sdg,
    rx,
    ry,
    rz,
    crz,
    ch,
    toffoli,
    reset,
    quantum_functional,
    project_z,
)


def compile_quantum_guppy(fn) -> ModulePointer:
    """A decorator that combines @guppy with HUGR compilation.

    Modified version of `tests.util.compile_guppy` that loads the quantum module.
    """
    assert not isinstance(
        fn,
        GuppyModule,
    ), "`@compile_quantum_guppy` does not support extra arguments."

    module = GuppyModule("module")
    module.load(
        angle, qubit, discard, measure, measure_array, maybe_qubit, discard_array
    )
    module.load(q=quantum)
    module.load_all(quantum_functional)
    guppylang.decorator.guppy(module)(fn)
    return module.compile()


def test_alloc(validate):
    @compile_quantum_guppy
    def test() -> tuple[bool, bool]:
        q1, q2 = qubit(), maybe_qubit().unwrap()
        q1, q2 = cx(q1, q2)
        return (measure(q1), measure(q2))

    validate(test)


def test_1qb_op(validate):
    @compile_quantum_guppy
    def test(q: qubit @ owned) -> qubit:
        q = h(q)
        q = t(q)
        q = s(q)
        q = x(q)
        q = y(q)
        q = z(q)
        q = tdg(q)
        q = sdg(q)
        return q

    validate(test)


def test_2qb_op(validate):
    @compile_quantum_guppy
    def test(q1: qubit @ owned, q2: qubit @ owned) -> tuple[qubit, qubit]:
        q1, q2 = cx(q1, q2)
        q1, q2 = cy(q1, q2)
        q1, q2 = cz(q1, q2)
        q1, q2 = ch(q1, q2)
        return (q1, q2)

    validate(test)


def test_3qb_op(validate):
    @compile_quantum_guppy
    def test(
        q1: qubit @ owned, q2: qubit @ owned, q3: qubit @ owned
    ) -> tuple[qubit, qubit, qubit]:
        q1, q2, q3 = toffoli(q1, q2, q3)
        return (q1, q2, q3)

    validate(test)


def test_measure_ops(validate):
    """Compile various measurement-related operations."""

    @compile_quantum_guppy
    def test(q1: qubit @ owned, q2: qubit @ owned) -> tuple[bool, bool]:
        q1, b1 = project_z(q1)
        q1 = discard(q1)
        q2 = reset(q2)
        b2 = measure(q2)
        return (b1, b2)

    validate(test)


def test_parametric(validate):
    """Compile various parametric operations."""

    @compile_quantum_guppy
    def test(
        q1: qubit @ owned, q2: qubit @ owned, a1: angle, a2: angle, a3: angle
    ) -> tuple[qubit, qubit]:
        q1 = rx(q1, a1)
        q1 = ry(q1, a1)
        q2 = rz(q2, a3)
        q1, q2 = crz(q1, q2, a3)
        return (q1, q2)


def test_measure_array(validate):
    """Build and measure array."""

    @compile_quantum_guppy
    def test() -> array[bool, 10]:
        qs = array(qubit() for _ in range(10))
        return measure_array(qs)

    validate(test)


def test_discard_array(validate):
    """Build and discard array."""

    @compile_quantum_guppy
    def test() -> None:
        qs = array(qubit() for _ in range(10))
        discard_array(qs)

    validate(test)


def test_panic_discard(validate):
    """Panic while discarding qubit."""

    @compile_quantum_guppy
    @no_type_check
    def test() -> None:
        q = qubit()
        panic("I panicked!", q)

    validate(test)


def test_barrier(validate):
    """Barrier between ops."""

    @compile_quantum_guppy
    @no_type_check
    def test() -> None:
        q1, q2, q3, q4 = qubit(), qubit(), qubit(), qubit()

        q.h(q1)
        q.h(q2)
        barrier(q1, q2, q3)
        q.h(q3)

        q.cx(q1, q2)
        barrier(q2, q3)
        q.cx(q3, q4)

        discard(q1)
        discard(q2)
        barrier()  # does nothing
        discard(q3)
        discard(q4)

    validate(test)


def test_barrier_array(validate):
    """Barrier on array/struct access."""

    @compile_quantum_guppy
    @no_type_check
    def test() -> None:
        qs = array(qubit() for _ in range(4))
        q.h(qs[1])
        q.h(qs[2])
        barrier(qs[1], qs[2], qs[3])
        barrier(qs[1])
        q.h(qs[3])

        q.cx(qs[1], qs[2])
        barrier(qs[2], qs[3])
        q.cx(qs[3], qs[4])
        barrier(qs)
        discard_array(qs)

    validate(test)


def test_barrier_struct(validate):
    """Barrier on array/struct access."""

    module = GuppyModule("module")
    module.load(qubit, discard)
    module.load(q=quantum)

    @guppylang.decorator.guppy.struct(module)
    class S:
        q1: qubit
        q2: qubit
        q3: qubit
        q4: qubit

    @guppylang.decorator.guppy(module)
    @no_type_check
    def test() -> None:
        qs = S(qubit(), qubit(), qubit(), qubit())
        q.h(qs.q1)
        q.h(qs.q2)
        barrier(qs.q1, qs.q2, qs.q3)
        barrier(qs.q1)
        q.h(qs.q3)

        q.cx(qs.q1, qs.q2)
        barrier(qs.q2, qs.q3)
        q.cx(qs.q3, qs.q4)

        discard(qs.q1)
        discard(qs.q2)
        discard(qs.q3)
        discard(qs.q4)

    validate(module.compile())


def test_barrier_misc(validate):
    """Barrier on classical and non-place."""

    @compile_quantum_guppy
    @no_type_check
    def test() -> None:
        q1 = qubit()
        q.h(q1)
        x = 1
        barrier(q1, array(1, 2, 3), 2 + 3, x)

        result("c", x)
        result("c2", measure(q1))

    validate(test)



from typing import no_type_check

from guppylang.decorator import guppy
from guppylang.std.builtins import array, py
from guppylang.std.quantum import cx, discard_array, h, measure, project_z, qubit


@guppy
@no_type_check
def many_ctrl_flow() -> None:
    qs = array(qubit() for _ in range(py(20)))

    for i in range(py(20)):
        h(qs[i])

    [q0, *qs0] = qs
    if measure(q0):
        [q1, *qs1] = qs0
        if measure(q1):
            [q2, *qs2] = qs1
            if measure(q2):
                [q3, *qs3] = qs2
                if measure(q3):
                    [q4, *qs4] = qs3
                    if measure(q4):
                        cx(qs4[0], qs4[1])
                    discard_array(qs4)
                else:
                    discard_array(qs3)
            else:
                discard_array(qs2)
        else:
            discard_array(qs1)
    else:
        cx(qs0[1], qs0[2])
        while project_z(qs0[0]):
            while project_z(qs0[1]):
                while project_z(qs0[2]):
                    while project_z(qs0[3]):
                        while project_z(qs0[4]):
                            while project_z(qs0[5]):
                                cx(qs0[1], qs0[2])
        discard_array(qs0)


def test_compile_many_ctrl_flow(benchmark):
    def comp() -> None:
        many_ctrl_flow.compile()

    benchmark(comp)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar

from guppylang.diagnostic import Error, Help, Note

if TYPE_CHECKING:
    from guppylang.definition.struct import StructField
    from guppylang.tys.const import Const
    from guppylang.tys.param import Parameter
    from guppylang.tys.ty import FunctionType, Type


@dataclass(frozen=True)
class TypeMismatchError(Error):
    title: ClassVar[str] = "Type mismatch"
    span_label: ClassVar[str] = "Expected {kind} of type `{expected}`, got `{actual}`"

    expected: Type
    actual: Type
    kind: str = "expression"

    @dataclass(frozen=True)
    class CantInferParam(Note):
        message: ClassVar[str] = (
            "Couldn't infer an instantiation for type variable `?{type_var}` "
            "(higher-rank polymorphic types are not supported)"
        )
        type_var: str

    @dataclass(frozen=True)
    class CantInstantiateFreeVars(Note):
        message: ClassVar[str] = (
            "Can't instantiate parameter `{param}` with type `{illegal_inst}` "
            "containing free variables"
        )
        param: str
        illegal_inst: Type | Const


@dataclass(frozen=True)
class AssignFieldTypeMismatchError(Error):
    title: ClassVar[str] = "Type mismatch"
    span_label: ClassVar[str] = (
        "Cannot assign expression of type `{actual}` to field `{field.name}` of type "
        "`{field.ty}`"
    )
    actual: Type
    field: StructField


@dataclass(frozen=True)
class AssignSubscriptTypeMismatchError(Error):
    title: ClassVar[str] = "Type mismatch"
    span_label: ClassVar[str] = (
        "Cannot assign expression of type `{actual}` to array element of type "
        "`{expected}`"
    )
    actual: Type
    expected: Type


@dataclass(frozen=True)
class NonLinearInstantiateError(Error):
    title: ClassVar[str] = "Not defined for linear argument"
    span_label: ClassVar[str] = (
        "Cannot instantiate non-linear type parameter `{param.name}` in type "
        "`{func_ty}` with linear type `{ty}`"
    )
    param: Parameter
    func_ty: FunctionType
    ty: Type


@dataclass(frozen=True)
class TypeInferenceError(Error):
    title: ClassVar[str] = "Cannot infer type"
    span_label: ClassVar[str] = (
        "Cannot infer type variables in expression of type `{unsolved_ty}`"
    )
    unsolved_ty: Type


@dataclass(frozen=True)
class IllegalConstant(Error):
    title: ClassVar[str] = "Unsupported constant"
    span_label: ClassVar[str] = "Type `{python_ty.__name__}` is not supported"
    python_ty: type


@dataclass(frozen=True)
class ModuleMemberNotFoundError(Error):
    title: ClassVar[str] = "Not found in module"
    span_label: ClassVar[str] = "Module `{module_name}` has no member `{member}`"
    module_name: str
    member: str


@dataclass(frozen=True)
class AttributeNotFoundError(Error):
    title: ClassVar[str] = "Attribute not found"
    span_label: ClassVar[str] = "`{ty}` has no attribute `{attribute}`"
    ty: Type
    attribute: str


@dataclass(frozen=True)
class UnaryOperatorNotDefinedError(Error):
    title: ClassVar[str] = "Operator not defined"
    span_label: ClassVar[str] = "Unary operator `{op}` not defined for `{ty}`"
    ty: Type
    op: str


@dataclass(frozen=True)
class BinaryOperatorNotDefinedError(Error):
    title: ClassVar[str] = "Operator not defined"
    span_label: ClassVar[str] = (
        "Binary operator `{op}` not defined for `{left_ty}` and `{right_ty}`"
    )
    left_ty: Type
    right_ty: Type
    op: str


@dataclass(frozen=True)
class BadProtocolError(Error):
    title: ClassVar[str] = "Not {is_not}"
    span_label: ClassVar[str] = "Expression of type `{ty}` is not {is_not}"
    ty: Type
    is_not: str

    @dataclass(frozen=True)
    class MethodMissing(Help):
        message: ClassVar[str] = "Implement missing method: `{method}: {signature}`"
        method: str
        signature: FunctionType

    @dataclass(frozen=True)
    class BadSignature(Help):
        message: ClassVar[str] = (
            "Fix signature of method `{ty}.{method}`:  Expected `{exp_signature}`, got "
            "`{act_signature}`"
        )
        ty: Type
        method: str
        exp_signature: FunctionType
        act_signature: FunctionType


@dataclass(frozen=True)
class MissingReturnValueError(Error):
    title: ClassVar[str] = "Missing return value"
    span_label: ClassVar[str] = "Expected return value of type `{ty}`"
    ty: Type


@dataclass(frozen=True)
class TypeApplyNotGenericError(Error):
    title: ClassVar[str] = "Not generic"
    span_label: ClassVar[str] = (
        "{thing} is not generic, so no type parameters can be provided"
    )
    func_name: str | None

    @property
    def thing(self) -> str:
        return f"`{self.func_name}`" if self.func_name else "This function"


@dataclass(frozen=True)
class NotCallableError(Error):
    title: ClassVar[str] = "Not callable"
    span_label: ClassVar[str] = "Expected a function, got expression of type `{actual}`"
    actual: Type


@dataclass(frozen=True)
class WrongNumberOfArgsError(Error):
    title: ClassVar[str] = ""  # Custom implementation in `rendered_title`
    expected: int
    actual: int
    detailed: bool = True
    is_type_apply: bool = False

    @property
    def rendered_title(self) -> str:
        return (
            f"Not enough {self.argument_kind}s"
            if self.expected > self.actual
            else f"Too many {self.argument_kind}s"
        )

    @property
    def argument_kind(self) -> str:
        return "type argument" if self.is_type_apply else "argument"

    @property
    def rendered_span_label(self) -> str:
        if not self.detailed:
            return f"Expected {self.expected}, got {self.actual}"
        diff = self.expected - self.actual
        if diff < 0:
            msg = f"Unexpected {self.argument_kind}"
            if diff < -1:
                msg += "s"
        else:
            msg = f"Missing {self.argument_kind}"
            if diff > 1:
                msg += "s"
        return f"{msg} (expected {self.expected}, got {self.actual})"

    @dataclass(frozen=True)
    class SignatureHint(Note):
        message: ClassVar[str] = "Function signature is `{sig}`"
        sig: FunctionType


@dataclass(frozen=True)
class WrongNumberOfUnpacksError(Error):
    title: ClassVar[str] = "{prefix} values to unpack"
    expected: int
    actual: int
    at_least: bool

    @property
    def prefix(self) -> str:
        return "Not enough" if self.expected > self.actual else "Too many"

    @property
    def rendered_span_label(self) -> str:
        diff = self.expected - self.actual
        if diff < 0:
            msg = "Unexpected assignment " + ("targets" if diff < -1 else "target")
            at_least = "at least " if self.at_least else ""
        else:
            msg = "Not enough assignment targets"
            assert not self.at_least
            at_least = ""
        return f"{msg} (expected {self.expected}, got {at_least}{self.actual})"


@dataclass(frozen=True)
class UnpackableError(Error):
    title: ClassVar[str] = "Unpackable"
    span_label: ClassVar[str] = "Expression of type `{ty}` cannot be unpacked"
    ty: Type

    @dataclass(frozen=True)
    class NonStaticIter(Note):
        message: ClassVar[str] = (
            "Unpacking of iterable types like `{ty}` is only allowed if the number of "
            "items yielded by the iterator is statically known. This is not the case "
            "for `{ty}`."
        )

    @dataclass(frozen=True)
    class GenericSize(Note):
        message: ClassVar[str] = (
            "Unpacking of iterable types like `{ty}` is only allowed if the number of "
            "items yielded by the iterator is statically known. Here, the number of "
            "items `{num}` is generic and can change between different function "
            "invocations."
        )
        num: Const


@dataclass(frozen=True)
class StarredTupleUnpackError(Error):
    title: ClassVar[str] = "Invalid starred unpacking"
    span_label: ClassVar[str] = (
        "Expression of type `{ty}` cannot be collected into a starred assignment since "
        "the yielded items have different types"
    )
    ty: Type


@dataclass(frozen=True)
class AssignNonPlaceHelp(Help):
    message: ClassVar[str] = (
        "Consider assigning this value to a local variable first before assigning the "
        "field `{field.name}`"
    )
    field: StructField


@dataclass(frozen=True)
class ArrayComprUnknownSizeError(Error):
    title: ClassVar[str] = "Array comprehension with nonstatic size"
    span_label: ClassVar[str] = "Cannot infer the size of this array comprehension ..."

    @dataclass(frozen=True)
    class IfGuard(Note):
        span_label: ClassVar[str] = "since it depends on this condition"

    @dataclass(frozen=True)
    class DynamicIterator(Note):
        span_label: ClassVar[str] = (
            "since the number of elements yielded by this iterator is not statically "
            "known"
        )



"""Dummy module used in `test_imports.py`"""

import hugr.tys as ht

from guppylang import GuppyModule, guppy

mod_a = GuppyModule("mod_a")


@guppy(mod_a)
def f(x: int) -> int:
    return x + 1


@guppy.declare(mod_a)
def g() -> int: ...


@guppy.type(ht.Bool, module=mod_a)
class MyType:
    @guppy.declare(mod_a)
    def __neg__(self: "MyType") -> "MyType": ...



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit
from guppylang.std.quantum_functional import h

module = GuppyModule("test")
module.load(qubit, h)


@guppy(module)
def test(n: int) -> None:
    q = qubit()
    i = 0
    while i < n:
        q = h(q)
        i += 1


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array

module = GuppyModule("test")


@guppy(module)
def foo(xs: array[int, 42]) -> int:
    return xs[1.0]


module.compile()



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load_all(quantum)


@guppy.struct(module)
class MyStruct:
    q: qubit


@guppy(module)
def foo(b: bool, s: MyStruct @owned) -> MyStruct:
    if b:
        s.q = qubit()
    else:
        s.q = qubit()
    s.q = qubit()
    return s


module.compile()



import importlib.util
import pathlib
import pytest
import sys
from hugr import tys
from hugr.tys import TypeBound

from guppylang.error import GuppyError
from guppylang.module import GuppyModule

import guppylang.decorator as decorator


def run_error_test(file, capsys, snapshot):
    file = pathlib.Path(file)

    with pytest.raises(GuppyError) as exc_info:
        importlib.import_module(f"tests.error.{file.parent.name}.{file.name}")

    # Invoke except hook to print the exception to stderr
    sys.excepthook(exc_info.type, exc_info.value, exc_info.tb)

    err = capsys.readouterr().err
    err = err.replace(str(file), "$FILE")

    snapshot.snapshot_dir = str(file.parent)
    snapshot.assert_match(err, file.with_suffix(".err").name)


util = GuppyModule("test")


@decorator.guppy.type(
    tys.Opaque(extension="", id="", args=[], bound=TypeBound.Copyable), module=util
)
class NonBool:
    pass



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned

module = GuppyModule("test")
module.load_all(quantum)


@guppy(module)
def foo(qs: list[qubit] @owned, xs: list[int]) -> list[qubit]:
    return [q for q in qs for x in xs]


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule

module = GuppyModule("test")


@guppy.struct(module)
class MyStruct:
    x: int
    y: bool


@guppy(module)
def foo(s: MyStruct) -> None:
    s.x, a = (1, 2), 3


module.compile()



from typing import no_type_check
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.option import Option, nothing, some
from guppylang.std.quantum import qubit

def test_none(validate, run_int_fn):
    module = GuppyModule("test_range")

    @guppy(module)
    @no_type_check
    def main() -> int:
        x: Option[int] = nothing()
        is_none = 10 if x.is_nothing() else 0
        is_some = 1 if x.is_some() else 0
        return is_none + is_some

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=10)


def test_some_unwrap(validate, run_int_fn):
    module = GuppyModule("test_range")

    @guppy(module)
    @no_type_check
    def main() -> int:
        x: Option[int] = some(42)
        is_none = 1 if x.is_nothing() else 0
        is_some = x.unwrap() if x.is_some() else 0
        return is_none + is_some

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=42)

def test_nothing_unwrap(validate, run_int_fn):
    module = GuppyModule("test_range")
    module.load(qubit)

    @guppy(module)
    @no_type_check
    def main() -> int:
        x: Option[qubit] = nothing()
        x.unwrap_nothing() # linearity error without this line
        return 1

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=1)


def test_take(validate, run_int_fn):
    module = GuppyModule("test_range")

    @guppy(module)
    @no_type_check
    def main() -> int:
        x: Option[int] = some(42)
        y = x.take().unwrap()
        is_none = 1 if x.is_nothing() else 0
        return y + is_none

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=43)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, TypeAlias

from hugr import tys as ht

from guppylang.error import InternalGuppyError
from guppylang.tys.common import ToHugr, Transformable, Transformer, Visitor
from guppylang.tys.const import BoundConstVar, Const, ConstValue, ExistentialConstVar
from guppylang.tys.var import ExistentialVar

if TYPE_CHECKING:
    from guppylang.tys.ty import Type


# We define the `Argument` type as a union of all `ArgumentBase` subclasses defined
# below. This models an algebraic data type and enables exhaustiveness checking in
# pattern matches etc.
# Note that this might become obsolete in case the `@sealed` decorator is added:
#  * https://peps.python.org/pep-0622/#sealed-classes-as-algebraic-data-types
#  * https://github.com/johnthagen/sealed-typing-pep
Argument: TypeAlias = "TypeArg | ConstArg"


@dataclass(frozen=True)
class ArgumentBase(ToHugr[ht.TypeArg], Transformable["Argument"], ABC):
    """Abstract base class for arguments of parametrized types.

    For example, in the type `array[int, 42]` we have two arguments `int` and `42`.
    """

    @property
    @abstractmethod
    def unsolved_vars(self) -> set[ExistentialVar]:
        """The existential type variables contained in this argument."""


@dataclass(frozen=True)
class TypeArg(ArgumentBase):
    """Argument that can be instantiated for a `TypeParameter`."""

    # The type to instantiate
    ty: "Type"

    @property
    def unsolved_vars(self) -> set[ExistentialVar]:
        """The existential type variables contained in this argument."""
        return self.ty.unsolved_vars

    def to_hugr(self) -> ht.TypeTypeArg:
        """Computes the Hugr representation of the argument."""
        ty: ht.Type = self.ty.to_hugr()
        return ty.type_arg()

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this argument."""
        if not visitor.visit(self):
            self.ty.visit(visitor)

    def transform(self, transformer: Transformer) -> Argument:
        """Accepts a transformer on this argument."""
        return transformer.transform(self) or TypeArg(self.ty.transform(transformer))


@dataclass(frozen=True)
class ConstArg(ArgumentBase):
    """Argument that can be substituted for a `ConstParam`."""

    const: Const

    @property
    def unsolved_vars(self) -> set[ExistentialVar]:
        """The existential const variables contained in this argument."""
        return self.const.unsolved_vars

    def to_hugr(self) -> ht.TypeArg:
        """Computes the Hugr representation of this argument."""
        from guppylang.tys.ty import NumericType

        match self.const:
            case ConstValue(value=v, ty=NumericType(kind=NumericType.Kind.Nat)):
                assert isinstance(v, int)
                return ht.BoundedNatArg(n=v)
            case BoundConstVar(idx=idx, ty=NumericType(kind=NumericType.Kind.Nat)):
                param = ht.BoundedNatParam(upper_bound=None)
                return ht.VariableArg(idx=idx, param=param)
            case ConstValue() | BoundConstVar():
                # TODO: Handle other cases besides nats
                raise NotImplementedError
            case ExistentialConstVar():
                raise InternalGuppyError(
                    "Tried to convert unsolved constant variable to Hugr"
                )

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this argument."""
        visitor.visit(self)

    def transform(self, transformer: Transformer) -> Argument:
        """Accepts a transformer on this argument."""
        return transformer.transform(self) or ConstArg(
            self.const.transform(transformer)
        )



"""Tests validating the files in the `examples` directory."""


def test_demo_notebook(nb_regression):
    nb_regression.diff_ignore += ("/metadata/language_info/version",)
    nb_regression.check("tests/integration/notebooks/demo.ipynb")


def test_random_walk_qpe(validate):
    from examples.random_walk_qpe import hugr

    validate(hugr)


def test_t_factory(validate):
    from examples.t_factory import hugr

    validate(hugr)



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")

T = guppy.type_var("T", module=module)


@guppy(module)
def foo(x: T) -> None:
    y = T


module.compile()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, TypeAlias

from guppylang.error import InternalGuppyError
from guppylang.tys.common import Transformable, Transformer, Visitor
from guppylang.tys.var import BoundVar, ExistentialVar

if TYPE_CHECKING:
    from guppylang.tys.arg import ConstArg
    from guppylang.tys.ty import Type


@dataclass(frozen=True)
class ConstBase(Transformable["Const"], ABC):
    """Abstract base class for constants arguments in the type system.

    In principle, we can allow constants of any type representable in the type system.
    For now, we will only support basic consts like `int` or `bool`, but in the future
    we could have struct constants etc.
    """

    ty: "Type"

    def __post_init__(self) -> None:
        if self.ty.unsolved_vars:
            raise InternalGuppyError("Attempted to create constant with unsolved type")

    @abstractmethod
    def cast(self) -> "Const":
        """Casts an implementor of `ConstBase` into a `Const`.

        This enforces that all implementors of `ConstBase` can be embedded into the
        `Const` union type.
        """

    @property
    def unsolved_vars(self) -> set[ExistentialVar]:
        """The existential type variables contained in this constant."""
        return set()

    def __str__(self) -> str:
        from guppylang.tys.printing import TypePrinter

        return TypePrinter().visit(self.cast())

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this constant."""
        visitor.visit(self)

    def transform(self, transformer: Transformer, /) -> "Const":
        """Accepts a transformer on this constant."""
        return transformer.transform(self) or self.cast()

    def to_arg(self) -> "ConstArg":
        """Wraps this constant into a type argument."""
        from guppylang.tys.arg import ConstArg

        return ConstArg(self.cast())


@dataclass(frozen=True)
class ConstValue(ConstBase):
    """A constant value in the type system.

    For example, in the type `array[int, 5]` the second argument is a `ConstArg`  that
    contains a `ConstValue(5)`.
    """

    # TODO: We will need a proper Guppy representation of this in the future
    value: Any

    def cast(self) -> "Const":
        """Casts an implementor of `ConstBase` into a `Const`."""
        return self


@dataclass(frozen=True)
class BoundConstVar(BoundVar, ConstBase):
    """Bound variable referencing a `ConstParam`.

    For example, in the function type `forall n: int. array[float, n] -> array[int, n]`,
    we represent the int argument to `array` as a `ConstArg` containing a
    `BoundConstVar(idx=0)`.
    """

    def cast(self) -> "Const":
        """Casts an implementor of `ConstBase` into a `Const`."""
        return self


@dataclass(frozen=True)
class ExistentialConstVar(ExistentialVar, ConstBase):
    """Existential constant variable.

    During type checking we try to solve all existential constant variables and
    substitute them with concrete constants.
    """

    @classmethod
    def fresh(cls, display_name: str, ty: "Type") -> "ExistentialConstVar":
        return ExistentialConstVar(ty, display_name, next(cls._fresh_id))

    @property
    def unsolved_vars(self) -> set[ExistentialVar]:
        """The existential type variables contained in this constant."""
        return {self}

    def cast(self) -> "Const":
        """Casts an implementor of `ConstBase` into a `Const`."""
        return self


Const: TypeAlias = ConstValue | BoundConstVar | ExistentialConstVar



from hugr import ops

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from tests.util import compile_guppy


def test_id(validate):
    @compile_guppy
    def identity(x: int) -> int:
        return x

    validate(identity)


def test_void(validate):
    @compile_guppy
    def void() -> None:
        return

    validate(void)


def test_copy(validate):
    @compile_guppy
    def copy(x: int) -> tuple[int, int]:
        return x, x

    validate(copy)


def test_discard(validate):
    @compile_guppy
    def discard(x: int) -> None:
        return

    validate(discard)


def test_implicit_return(validate):
    @compile_guppy
    def ret() -> None:
        pass

    validate(ret)


def test_assign(validate):
    @compile_guppy
    def foo(x: bool) -> bool:
        y = x
        return y

    validate(foo)


def test_assign_expr(validate):
    @compile_guppy
    def foo(x: bool) -> bool:
        (y := x)
        return y

    validate(foo)


def test_func_def_name():
    @compile_guppy
    def func_name() -> None:
        return

    assert func_name.func_defn.f_name == "func_name"


def test_func_decl_name():
    module = GuppyModule("test")

    @guppy.declare(module)
    def func_name() -> None: ...

    hugr = module.compile_hugr()
    [def_op] = [
        data.op for n, data in hugr.nodes() if isinstance(data.op, ops.FuncDecl)
    ]
    assert isinstance(def_op, ops.FuncDecl)
    assert def_op.f_name == "func_name"


def test_compile_again():
    module = GuppyModule("test")

    @guppy(module)
    def identity(x: int) -> int:
        return x

    hugr = module.compile()

    # Compiling again should return the same Hugr
    assert hugr is module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit

module = GuppyModule("test")
module.load(qubit)


@guppy.declare(module)
def foo(q1: qubit, q2: qubit @owned) -> qubit: ...


@guppy(module)
def test(q1: qubit @owned, q2: qubit @owned) -> tuple[qubit, qubit]:
    q1 = foo(q1, q2)
    return q1, q2


module.compile()



"""Compilers building list functions on top of hugr standard operations, that involve
multiple nodes.
"""

from __future__ import annotations

from hugr import Wire, ops
from hugr import ext as he
from hugr import tys as ht
from hugr.std.float import FLOAT_T
from tket2_exts import (
    futures,
    qsystem,
    qsystem_random,
    qsystem_utils,
    quantum,
    result,
    rotation,
)

from guppylang.definition.custom import CustomInoutCallCompiler
from guppylang.definition.value import CallReturnWires

# ----------------------------------------------
# --------- tket2.* extensions -----------------
# ----------------------------------------------

FUTURES_EXTENSION = futures()
QSYSTEM_EXTENSION = qsystem()
QSYSTEM_RANDOM_EXTENSION = qsystem_random()
QSYSTEM_UTILS_EXTENSION = qsystem_utils()
QUANTUM_EXTENSION = quantum()
RESULT_EXTENSION = result()
ROTATION_EXTENSION = rotation()

RNGCONTEXT_T_DEF = QSYSTEM_RANDOM_EXTENSION.get_type("context")
RNGCONTEXT_T = ht.ExtType(RNGCONTEXT_T_DEF)

ROTATION_T_DEF = ROTATION_EXTENSION.get_type("rotation")
ROTATION_T = ht.ExtType(ROTATION_T_DEF)

TKET2_EXTENSIONS = [
    FUTURES_EXTENSION,
    QSYSTEM_EXTENSION,
    QSYSTEM_RANDOM_EXTENSION,
    QSYSTEM_UTILS_EXTENSION,
    QUANTUM_EXTENSION,
    RESULT_EXTENSION,
    ROTATION_EXTENSION,
]


def from_halfturns_unchecked() -> ops.ExtOp:
    return ops.ExtOp(
        ROTATION_EXTENSION.get_op("from_halfturns_unchecked"),
        ht.FunctionType([FLOAT_T], [ROTATION_T]),
    )


# ------------------------------------------------------
# --------- Custom compilers for non-native ops --------
# ------------------------------------------------------


class InoutMeasureCompiler(CustomInoutCallCompiler):
    """Compiler for the measure functions with an inout qubit
    such as the `project_z` function."""

    opname: str
    ext: he.Extension

    def __init__(self, opname: str | None = None, ext: he.Extension | None = None):
        self.opname = opname or "Measure"
        self.ext = ext or QUANTUM_EXTENSION

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        from guppylang.std._internal.util import quantum_op

        [q] = args
        [q, bit] = self.builder.add_op(
            quantum_op(self.opname, ext=self.ext)(
                ht.FunctionType([ht.Qubit], [ht.Qubit, ht.Bool]), []
            ),
            q,
        )
        return CallReturnWires(regular_returns=[bit], inout_returns=[q])


class RotationCompiler(CustomInoutCallCompiler):
    opname: str

    def __init__(self, opname: str):
        self.opname = opname

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        from guppylang.std._internal.util import quantum_op

        [*qs, angle] = args
        [halfturns] = self.builder.add_op(ops.UnpackTuple([FLOAT_T]), angle)
        [rotation] = self.builder.add_op(from_halfturns_unchecked(), halfturns)

        qs = self.builder.add_op(
            quantum_op(self.opname)(
                ht.FunctionType(
                    [ht.Qubit for _ in qs] + [ROTATION_T], [ht.Qubit for _ in qs]
                ),
                [],
            ),
            *qs,
            rotation,
        )
        return CallReturnWires(regular_returns=[], inout_returns=list(qs))



from tests.util import compile_guppy


def test_basic(validate):
    @compile_guppy
    def foo(x: int) -> int:
        def bar(y: int) -> int:
            return y

        return bar(x + 1)

    validate(foo)


def test_call_twice(validate):
    @compile_guppy
    def foo(x: int) -> int:
        def bar(y: int) -> int:
            return y + 3

        if x > 5:
            return bar(x)
        else:
            return bar(2 * x)

    validate(foo)


def test_redefine(validate):
    @compile_guppy
    def foo(x: int) -> int:
        def bar(y: int) -> int:
            return y + 3

        a = bar(x)

        def bar(y: int) -> int:
            return y

        b = bar(0)
        return a + b

    validate(foo)


def test_define_twice(validate):
    @compile_guppy
    def foo(x: int) -> int:
        if x == 0:

            def bar(y: int) -> int:
                return y + 3
        else:

            def bar(y: int) -> int:
                return y - 42

        return bar(x)

    validate(foo)


def test_nested_deep(validate):
    @compile_guppy
    def foo(x: int) -> int:
        def bar(y: int) -> int:
            def baz(z: int) -> int:
                return z - 1

            return baz(5 * y)

        return bar(x + 1)

    validate(foo)


def test_recurse(validate):
    @compile_guppy
    def foo(x: int) -> int:
        def bar(y: int) -> int:
            if y == 0:
                return 0
            return 2 * bar(y - 1)

        return bar(x)

    validate(foo)


def test_capture_arg(validate):
    @compile_guppy
    def foo(x: int) -> int:
        def bar() -> int:
            return 1 + x

        return bar()

    validate(foo)


def test_capture_assigned(validate):
    @compile_guppy
    def foo(x: int) -> int:
        y = x + 1

        def bar() -> int:
            return y

        return bar()

    validate(foo)


def test_capture_multiple(validate):
    @compile_guppy
    def foo(x: int) -> int:
        if x > 5:
            y = 3
        else:
            y = 2 * x
        z = x + y

        def bar() -> int:
            q = y
            return q + z

        return bar()

    validate(foo)


def test_capture_fn(validate):
    @compile_guppy
    def foo() -> bool:
        def f(x: bool) -> bool:
            return x

        def g(b: bool) -> bool:
            return f(b)

        return g(True)

    validate(foo)


def test_capture_cfg(validate):
    @compile_guppy
    def foo(x: int) -> int:
        a = x + 4
        if x > 5:
            y = 5

            def bar() -> int:
                return x + y + a

            return bar()
        return 4

    validate(foo)


def test_capture_deep(validate):
    @compile_guppy
    def foo(x: int) -> int:
        a = x * 2

        def bar() -> int:
            b = a + 1

            def baz(y: int) -> int:
                c = a + b + y + x
                return c

            return baz(b * a)

        return bar()

    validate(foo)


def test_capture_recurse(validate):
    @compile_guppy
    def foo(x: int) -> int:
        def bar(y: int, z: int) -> int:
            if y == 0:
                return z
            return bar(z, z * x)

        return bar(x, 0)

    validate(foo)


def test_capture_recurse_nested(validate):
    @compile_guppy
    def foo(x: int) -> int:
        def bar(y: int, z: int) -> int:
            if y == 0:
                return z

            def baz() -> int:
                if z < 42:
                    return bar(z, z * x)
                return foo(z - 2)

            return baz()

        return bar(x, 0)

    validate(foo)


def test_capture_while(validate):
    @compile_guppy
    def foo(x: int) -> int:
        a = 0
        while x > 0:

            def bar() -> int:
                return x * x

            a += bar()
            x -= 1
        return a

    validate(foo)



import ast
from dataclasses import dataclass, field

from hugr import Node, Wire
from hugr import val as hv
from hugr.build.dfg import DefinitionBuilder, OpVar

from guppylang.ast_util import AstNode
from guppylang.checker.core import Globals
from guppylang.compiler.core import CompilerContext, DFContainer
from guppylang.definition.common import CompilableDef, ParsableDef
from guppylang.definition.value import CompiledValueDef, ValueDef
from guppylang.span import SourceMap
from guppylang.tys.parsing import type_from_ast


@dataclass(frozen=True)
class RawConstDef(ParsableDef):
    """A raw constant definition as provided by the user."""

    type_ast: ast.expr
    value: hv.Value

    description: str = field(default="constant", init=False)

    def parse(self, globals: Globals, sources: SourceMap) -> "ConstDef":
        """Parses and checks the user-provided signature of the function."""
        return ConstDef(
            self.id,
            self.name,
            self.defined_at,
            type_from_ast(self.type_ast, globals, {}),
            self.type_ast,
            self.value,
        )


@dataclass(frozen=True)
class ConstDef(RawConstDef, ValueDef, CompilableDef):
    """A constant with a checked type."""

    def compile_outer(self, graph: DefinitionBuilder[OpVar]) -> "CompiledConstDef":
        const_node = graph.add_const(self.value)
        return CompiledConstDef(
            self.id,
            self.name,
            self.defined_at,
            self.ty,
            self.type_ast,
            self.value,
            const_node,
        )


@dataclass(frozen=True)
class CompiledConstDef(ConstDef, CompiledValueDef):
    """A constant that has been compiled to a Hugr node."""

    const_node: Node

    def load(self, dfg: DFContainer, ctx: CompilerContext, node: AstNode) -> Wire:
        """Loads the extern value into a local Hugr dataflow graph."""
        return dfg.builder.load(self.const_node)



"""Source spans representing locations in the code being compiled."""

import ast
import linecache
from dataclasses import dataclass
from typing import TypeAlias

from guppylang.ast_util import get_file, get_line_offset
from guppylang.error import InternalGuppyError


@dataclass(frozen=True, order=True)
class Loc:
    """A location in a source file."""

    file: str

    #: Line number starting at 1
    line: int

    #: Column number starting at 1
    column: int

    def __str__(self) -> str:
        """Returns the string representation of this source location."""
        return f"{self.file}:{self.line}:{self.column}"

    def shift_left(self, cols: int) -> "Loc":
        """Returns a new location shifted to left by the given number of columns."""
        assert self.column >= cols
        return Loc(self.file, self.line, self.column - cols)

    def shift_right(self, cols: int) -> "Loc":
        """Returns a new location shifted to right by the given number of columns."""
        return Loc(self.file, self.line, self.column + cols)


@dataclass(frozen=True)
class Span:
    """A continuous sequence of source code within a file."""

    #: Starting location of the span (inclusive)
    start: Loc

    # Ending location of the span (exclusive)
    end: Loc

    def __post_init__(self) -> None:
        if self.start.file != self.end.file:
            raise InternalGuppyError("Span: Source spans multiple files")
        if self.start > self.end:
            raise InternalGuppyError("Span: Start after end")

    def __contains__(self, x: "Span | Loc") -> bool:
        """Determines whether another span or location is completely contained in this
        span."""
        if self.file != x.file:
            return False
        if isinstance(x, Span):
            return self.start <= x.start <= self.end <= x.end
        return self.start <= x <= self.end

    def __and__(self, other: "Span") -> "Span | None":
        """Returns the intersection with the given span or `None` if they don't
        intersect."""
        if self.file != other.file:
            return None
        if self.start > other.end or other.start > self.end:
            return None
        return Span(max(self.start, other.start), min(self.end, other.end))

    def __len__(self) -> int:
        """Returns the length of a single-line span in columns.

        Querying the length of multiline spans raises an `InternalGuppyError`.
        """
        if self.is_multiline:
            raise InternalGuppyError("Span: Tried to compute length of multi-line span")
        return self.end.column - self.start.column

    @property
    def file(self) -> str:
        """The file containing this span."""
        return self.start.file

    @property
    def is_multiline(self) -> bool:
        """Whether this source sequence spans multiple lines."""
        return self.start.line != self.end.line

    def shift_left(self, cols: int) -> "Span":
        """Returns a new span that is shifted to the left by the given number of
        columns."""
        return Span(self.start.shift_left(cols), self.end.shift_left(cols))

    def shift_right(self, cols: int) -> "Span":
        """Returns a new span that is shifted to the right by the given number of
        columns."""
        return Span(self.start.shift_right(cols), self.end.shift_right(cols))


#: Objects in the compiler that are associated with a source span
ToSpan: TypeAlias = ast.AST | Span


def to_span(x: ToSpan) -> Span:
    """Extracts a source span from an object."""
    if isinstance(x, Span):
        return x
    file, line_offset = get_file(x), get_line_offset(x)
    assert file is not None
    assert line_offset is not None
    # x.lineno and line_offset both start at 1, so we have to subtract 1
    start = Loc(file, x.lineno + line_offset - 1, x.col_offset)
    end = Loc(
        file,
        (x.end_lineno or x.lineno) + line_offset - 1,
        x.end_col_offset or x.col_offset,
    )
    return Span(start, end)


#: List of source lines in a file
SourceLines: TypeAlias = list[str]


class SourceMap:
    """Map holding the source code for all files accessed by the compiler.

    Can be used to look up the source code associated with a span.
    """

    sources: dict[str, SourceLines]

    def __init__(self) -> None:
        self.sources = {}

    def add_file(self, file: str, content: str | None = None) -> None:
        """Registers a new source file."""
        if content is None:
            self.sources[file] = [line.rstrip() for line in linecache.getlines(file)]
        else:
            self.sources[file] = content.splitlines(keepends=False)

    def span_lines(self, span: Span, prefix_lines: int = 0) -> list[str]:
        return self.sources[span.file][
            span.start.line - prefix_lines - 1 : span.end.line
        ]



from guppylang.decorator import guppy
from guppylang.module import GuppyModule

import tests.error.import_errors.lib as lib

module = GuppyModule("test")
module.load(lib)


@guppy(module)
def main(x: int) -> int:
    return lib.bar(x)


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit

module = GuppyModule("test")
module.load(qubit)


@guppy.struct(module)
class MyStruct:
    x: int
    q: qubit


@guppy(module)
def foo(s: MyStruct) -> tuple[MyStruct, bool]:
    t = s
    t.x += 1
    # People would expect `t.x == s.x`, but we can't do that
    # using the current tuple semantics
    return t, t.x == s.x


module.compile()



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load_all(quantum)


@guppy.struct(module)
class MyStruct:
    q: qubit


@guppy(module)
def foo(s: MyStruct @owned) -> tuple[qubit, qubit]:
    t = s
    return s.q, t.q


module.compile()



from guppylang.decorator import guppy
from guppylang.experimental import enable_experimental_features
from guppylang.module import GuppyModule
from guppylang.std import builtins, quantum
from guppylang.std.builtins import Array, Bool, Float, Int, List, py
from guppylang.std.quantum import qubit

# This is updated by our release-please workflow, triggered by this
# annotation: x-release-please-version
__version__ = "0.17.0"



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load_all(quantum)


@guppy(module)
def foo(q: qubit @owned) -> qubit:
    def bar() -> qubit:
        return q

    return q


module.compile()



from ast import expr
from dataclasses import dataclass
from types import TracebackType
from typing import ClassVar

from guppylang.ast_util import AstNode
from guppylang.diagnostic import Error, Help
from guppylang.error import GuppyError

EXPERIMENTAL_FEATURES_ENABLED = False


class enable_experimental_features:
    """Enables experimental Guppy features.

    Can be used as a context manager to enable experimental features in a `with` block.
    """

    def __init__(self) -> None:
        global EXPERIMENTAL_FEATURES_ENABLED
        self.original = EXPERIMENTAL_FEATURES_ENABLED
        EXPERIMENTAL_FEATURES_ENABLED = True

    def __enter__(self) -> None:
        pass

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        global EXPERIMENTAL_FEATURES_ENABLED
        EXPERIMENTAL_FEATURES_ENABLED = self.original


class disable_experimental_features:
    """Disables experimental Guppy features.

    Can be used as a context manager to enable experimental features in a `with` block.
    """

    def __init__(self) -> None:
        global EXPERIMENTAL_FEATURES_ENABLED
        self.original = EXPERIMENTAL_FEATURES_ENABLED
        EXPERIMENTAL_FEATURES_ENABLED = False

    def __enter__(self) -> None:
        pass

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        global EXPERIMENTAL_FEATURES_ENABLED
        EXPERIMENTAL_FEATURES_ENABLED = self.original


@dataclass(frozen=True)
class ExperimentalFeatureError(Error):
    title: ClassVar[str] = "Experimental feature"
    span_label: ClassVar[str] = "{things} are an experimental feature"
    things: str

    @dataclass(frozen=True)
    class Suggestion(Help):
        message: ClassVar[str] = (
            "Experimental features are currently disabled. You can enable them by "
            "calling `guppylang.enable_experimental_features()`, however note that "
            "these features are unstable and might break in the future."
        )

    def __post_init__(self) -> None:
        self.add_sub_diagnostic(ExperimentalFeatureError.Suggestion(None))


def check_function_tensors_enabled(node: expr | None = None) -> None:
    if not EXPERIMENTAL_FEATURES_ENABLED:
        raise GuppyError(ExperimentalFeatureError(node, "Function tensors"))


def check_lists_enabled(loc: AstNode | None = None) -> None:
    if not EXPERIMENTAL_FEATURES_ENABLED:
        raise GuppyError(ExperimentalFeatureError(loc, "Lists"))



from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from guppylang.definition.common import CompiledDef

if TYPE_CHECKING:
    from guppylang.checker.core import Globals


@dataclass(frozen=True)
class ModuleDef(CompiledDef):
    """A module definition.

    Note that this definition is separate from the `GuppyModule` class and only serves
    as a pointer to be stored in the globals.

    In the future we could consider unifying this with `GuppyModule`.
    """

    globals: "Globals"

    description: str = field(default="module", init=False)



"""Guppy standard module for quantum operations."""

from typing import no_type_check

import guppylang.std.angles as angles
import guppylang.std.quantum as quantum
from guppylang.decorator import guppy

# mypy: disable-error-code="empty-body, misc, valid-type"
from guppylang.module import GuppyModule
from guppylang.std.angles import angle
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit

quantum_functional = GuppyModule("quantum_functional")

quantum_functional.load(qubit, quantum)
quantum_functional.load_all(angles)


@guppy(quantum_functional)
@no_type_check
def h(q: qubit @ owned) -> qubit:
    quantum.h(q)
    return q


@guppy(quantum_functional)
@no_type_check
def cz(control: qubit @ owned, target: qubit @ owned) -> tuple[qubit, qubit]:
    quantum.cz(control, target)
    return control, target


@guppy(quantum_functional)
@no_type_check
def cx(control: qubit @ owned, target: qubit @ owned) -> tuple[qubit, qubit]:
    quantum.cx(control, target)
    return control, target


@guppy(quantum_functional)
@no_type_check
def cy(control: qubit @ owned, target: qubit @ owned) -> tuple[qubit, qubit]:
    quantum.cy(control, target)
    return control, target


@guppy(quantum_functional)
@no_type_check
def t(q: qubit @ owned) -> qubit:
    quantum.t(q)
    return q


@guppy(quantum_functional)
@no_type_check
def s(q: qubit @ owned) -> qubit:
    quantum.s(q)
    return q


@guppy(quantum_functional)
@no_type_check
def x(q: qubit @ owned) -> qubit:
    quantum.x(q)
    return q


@guppy(quantum_functional)
@no_type_check
def y(q: qubit @ owned) -> qubit:
    quantum.y(q)
    return q


@guppy(quantum_functional)
@no_type_check
def z(q: qubit @ owned) -> qubit:
    quantum.z(q)
    return q


@guppy(quantum_functional)
@no_type_check
def tdg(q: qubit @ owned) -> qubit:
    quantum.tdg(q)
    return q


@guppy(quantum_functional)
@no_type_check
def sdg(q: qubit @ owned) -> qubit:
    quantum.sdg(q)
    return q


@guppy(quantum_functional)
@no_type_check
def rz(q: qubit @ owned, angle: angle) -> qubit:
    quantum.rz(q, angle)
    return q


@guppy(quantum_functional)
@no_type_check
def rx(q: qubit @ owned, angle: angle) -> qubit:
    quantum.rx(q, angle)
    return q


@guppy(quantum_functional)
@no_type_check
def ry(q: qubit @ owned, angle: angle) -> qubit:
    quantum.ry(q, angle)
    return q


@guppy(quantum_functional)
@no_type_check
def crz(
    control: qubit @ owned, target: qubit @ owned, angle: angle
) -> tuple[qubit, qubit]:
    quantum.crz(control, target, angle)
    return control, target


@guppy(quantum_functional)
@no_type_check
def toffoli(
    control1: qubit @ owned, control2: qubit @ owned, target: qubit @ owned
) -> tuple[qubit, qubit, qubit]:
    quantum.toffoli(control1, control2, target)
    return control1, control2, target


@guppy(quantum_functional)
@no_type_check
def reset(q: qubit @ owned) -> qubit:
    quantum.reset(q)
    return q


@guppy(quantum_functional)
@no_type_check
def project_z(q: qubit @ owned) -> tuple[qubit, bool]:
    b = quantum.project_z(q)
    return q, b


# -------NON-PRIMITIVE-------


@guppy(quantum_functional)
@no_type_check
def ch(control: qubit @ owned, target: qubit @ owned) -> tuple[qubit, qubit]:
    quantum.ch(control, target)
    return control, target



"""Tools for inspecting source code when running in IPython."""

import ast
from typing import Any, NamedTuple, cast


def is_running_ipython() -> bool:
    """Checks if we are currently running in IPython"""
    try:
        return get_ipython() is not None  # type: ignore[name-defined]
    except NameError:
        return False


def is_ipython_dummy_file(filename: str) -> bool:
    """Checks whether a given filename is a dummy name generated for an IPython cell."""
    # TODO: The approach below is false-positive prone. Figure out if there is a better
    #  way to do this.
    return (
        # IPython cells have filenames like "<ipython-input-3-3e9b5833de21>"
        filename.startswith("<ipython-input-")
        # Jupyter cells have filenames like "/var/{...}/ipykernel_82076/61218616.py"
        or "ipykernel_" in filename
    )


def get_ipython_cell_sources() -> list[str]:
    """Returns the source code of all cells in the running IPython session.

    See https://github.com/wandb/weave/pull/1864
    """
    shell = get_ipython()  # type: ignore[name-defined]  # noqa: F821
    if not hasattr(shell, "user_ns"):
        raise AttributeError("Cannot access user namespace")
    cells = cast(list[str], shell.user_ns["In"])
    # First cell is always empty
    return cells[1:]


class IPythonDef(NamedTuple):
    """AST of a definition in IPython together with the definition cell name."""

    node: ast.FunctionDef | ast.ClassDef
    cell_name: str
    cell_source: str


def find_ipython_def(name: str) -> IPythonDef | None:
    """Tries to find a definition matching a given name in the current IPython session.

    Note that this only finds *top-level* function or class definitions. Nested
    definitions are not detected.

    See https://github.com/wandb/weave/pull/1864
    """
    cell_sources = get_ipython_cell_sources()
    # Search cells in reverse order to find the most recent version of the definition
    for i, cell_source in enumerate(reversed(cell_sources)):
        try:
            cell_ast = ast.parse(cell_source)
        except SyntaxError:
            continue
        # Search body in reverse order to find the most recent version of the class
        for node in reversed(cell_ast.body):
            if isinstance(node, ast.FunctionDef | ast.ClassDef) and node.name == name:
                cell_name = f"In [{len(cell_sources) - i}]"
                return IPythonDef(node, cell_name, cell_source)
    return None


def get_ipython_globals() -> dict[str, Any]:
    """Returns the globals of the current IPython kernel."""
    try:
        return get_ipython().user_ns  # type: ignore[name-defined, no-any-return]
    except NameError:
        raise RuntimeError("Not running in IPython") from None



from typing import no_type_check
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import barrier
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load(qubit)

@guppy(module)
@no_type_check
def main() -> qubit:
   q = qubit()
   barrier(q, q)
   return q


module.compile()



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load_all(quantum)


@guppy(module)
def main(qs: array[qubit, 42]) -> tuple[qubit, array[qubit, 42]]:
    q = qs[0]
    return q, qs


module.compile()


from guppylang import GuppyModule, guppy


def test_import_func(validate):
    from tests.integration.modules.mod_a import f, g

    module = GuppyModule("test")
    module.load(f, g)

    @guppy(module)
    def test(x: int) -> int:
        return f(x) + g()

    validate(module.compile())


def test_import_type(validate):
    from tests.integration.modules.mod_a import MyType

    module = GuppyModule("test")
    module.load(MyType)

    @guppy(module)
    def test(x: MyType) -> MyType:
        # Check that we've correctly imported the __neg__ method
        return -x

    validate(module.compile())


def test_import_implicit(validate):
    from tests.integration.modules.implicit_mod import foo

    module = GuppyModule("test")
    module.load(foo)

    @guppy(module)
    def test(x: int) -> int:
        return foo(x)

    validate(module.compile())


def test_func_alias(validate):
    from tests.integration.modules.mod_a import f as g

    module = GuppyModule("test")
    module.load(g=g)

    @guppy(module)
    def test(x: int) -> int:
        return g(x)

    validate(module.compile())


def test_type_alias(validate):
    from tests.integration.modules.mod_a import MyType

    module = GuppyModule("test")
    module.load(MyType_Alias=MyType)

    @guppy(module)
    def test(x: "MyType_Alias") -> "MyType_Alias":
        # Check that we still have the __neg__ method
        return -x

    validate(module.compile())


def test_conflict_alias(validate):
    from tests.integration.modules.mod_a import f
    from tests.integration.modules.mod_b import f as f_b

    module = GuppyModule("test")
    module.load(f, f_b=f_b)

    @guppy(module)
    def test(x: int, y: bool) -> tuple[int, bool]:
        return f(x), f_b(y)

    validate(module.compile())


def test_conflict_alias_type(validate):
    from tests.integration.modules.mod_a import MyType
    from tests.integration.modules.mod_b import MyType as MyTypeB

    module = GuppyModule("test")
    module.load(MyType, MyTypeB=MyTypeB)

    @guppy(module)
    def test(x: MyType, y: MyTypeB) -> tuple[MyType, MyTypeB]:
        return -x, +y

    validate(module.compile())


def test_type_transitive(validate):
    from tests.integration.modules.mod_c import g

    module = GuppyModule("test")
    module.load(g)  # `g` returns a type that was defined in `mod_a`

    @guppy(module)
    def test() -> int:
        x = g()  # Type of x was defined in `mod_a`
        return int(-x)  # Call method that was defined in `mod_a`

    validate(module.compile())


def test_type_transitive_conflict(validate):
    from tests.integration.modules.mod_b import MyType
    from tests.integration.modules.mod_c import g

    module = GuppyModule("test")
    module.load(MyType, g)

    @guppy(module)
    def test(ty_b: MyType) -> MyType:
        ty_a = g()  # g returns a type with the same name `MyType`
        return +ty_b

    validate(module.compile())


def test_func_transitive(validate):
    from tests.integration.modules.mod_c import h

    module = GuppyModule("test")
    module.load(h)  # `h` calls a function that was defined in `mod_a`

    @guppy(module)
    def test(x: int) -> int:
        return h(x)

    validate(module.compile())


def test_qualified(validate):
    import tests.integration.modules.mod_a as mod_a
    import tests.integration.modules.mod_b as mod_b

    module = GuppyModule("test")
    module.load(mod_a, mod_b)

    @guppy(module)
    def test(x: int, y: bool) -> tuple[int, bool]:
        return mod_a.f(x), mod_b.f(y)

    validate(module.compile())


def test_qualified_types(validate):
    import tests.integration.modules.mod_a as mod_a
    import tests.integration.modules.mod_b as mod_b

    module = GuppyModule("test")
    module.load(mod_a, mod_b)

    @guppy(module)
    def test(x: mod_a.MyType, y: mod_b.MyType) -> tuple[mod_a.MyType, mod_b.MyType]:
        return -x, +y

    validate(module.compile())


def test_qualified_implicit(validate):
    import tests.integration.modules.implicit_mod as implicit_mod

    module = GuppyModule("test")
    module.load(implicit_mod)

    @guppy(module)
    def test(x: int) -> int:
        return implicit_mod.foo(x)

    validate(module.compile())


def test_private_func(validate):
    # First, define a module with a public function
    # that calls an internal one
    internal_module = GuppyModule("test_internal")

    @guppy(internal_module)
    def _internal(x: int) -> int:
        return x

    @guppy(internal_module)
    def g(x: int) -> int:
        return _internal(x)

    # The test module
    module = GuppyModule("test")
    module.load_all(internal_module)

    @guppy(module)
    def f(x: int) -> int:
        return g(x)

    hugr = module.compile()
    validate(hugr)



from typing import Callable

from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")

@guppy.declare(module)
def foo(f: "Callable[int, float, bool]") -> None: ...


module.compile()



from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, TypeAlias

from hugr import tys as ht
from typing_extensions import Self

from guppylang.ast_util import AstNode
from guppylang.checker.errors.generic import ExpectedError
from guppylang.checker.errors.type_errors import TypeMismatchError
from guppylang.error import GuppyError, GuppyTypeError, InternalGuppyError
from guppylang.tys.arg import Argument, ConstArg, TypeArg
from guppylang.tys.common import ToHugr
from guppylang.tys.const import BoundConstVar, ExistentialConstVar
from guppylang.tys.errors import WrongNumberOfTypeArgsError
from guppylang.tys.var import ExistentialVar

if TYPE_CHECKING:
    from guppylang.tys.ty import Type


# We define the `Parameter` type as a union of all `ParameterBase` subclasses defined
# below. This models an algebraic data type and enables exhaustiveness checking in
# pattern matches etc.
# Note that this might become obsolete in case the `@sealed` decorator is added:
#  * https://peps.python.org/pep-0622/#sealed-classes-as-algebraic-data-types
#  * https://github.com/johnthagen/sealed-typing-pep
Parameter: TypeAlias = "TypeParam | ConstParam"


@dataclass(frozen=True)
class ParameterBase(ToHugr[ht.TypeParam], ABC):
    """Abstract base class for parameters used in function and type definitions.

    For example, when defining a struct type

    ```
    @guppy.struct
    class Foo[T, n: int]:
        ...
    ```

    we generate an `StructDef` that depends on the parameters `T` and `n`. From
    this, we obtain a proper `StructType` by providing arguments that are substituted
    for the parameters (for example `Foo[int, 42]`).
    """

    idx: int
    name: str

    @abstractmethod
    def with_idx(self, idx: int) -> Self:
        """Returns a copy of the parameter with a new index."""

    @abstractmethod
    def check_arg(self, arg: Argument, loc: AstNode | None = None) -> Argument:
        """Checks that this parameter can be instantiated with a given argument.

        Raises a user error if the argument is not valid.
        """

    @abstractmethod
    def to_existential(self) -> tuple[Argument, ExistentialVar]:
        """Creates a fresh existential variable that can be instantiated for this
        parameter.

        Returns both the argument and the created variable.
        """

    @abstractmethod
    def to_bound(self, idx: int | None = None) -> Argument:
        """Creates a bound variable with a given index that can be instantiated for this
        parameter.
        """


@dataclass(frozen=True)
class TypeParam(ParameterBase):
    """A parameter of kind type. Used to define generic functions and types."""

    must_be_copyable: bool
    must_be_droppable: bool

    @property
    def can_be_linear(self) -> bool:
        """Whether this type should be treated linearly."""
        return not self.must_be_copyable and not self.must_be_droppable

    def with_idx(self, idx: int) -> "TypeParam":
        """Returns a copy of the parameter with a new index."""
        return TypeParam(idx, self.name, self.must_be_copyable, self.must_be_droppable)

    def check_arg(self, arg: Argument, loc: AstNode | None = None) -> TypeArg:
        """Checks that this parameter can be instantiated with a given argument.

        Raises a user error if the argument is not valid.
        """
        match arg:
            case ConstArg(const):
                err = ExpectedError(loc, "a type", got=f"value of type `{const.ty}`")
                raise GuppyTypeError(err)
            case TypeArg(ty):
                if not self.can_be_linear and ty.linear:
                    err = ExpectedError(
                        loc, "a non-linear type", got=f"value of type `{ty}`"
                    )
                    raise GuppyTypeError(err)
                return arg

    def to_existential(self) -> tuple[Argument, ExistentialVar]:
        """Creates a fresh existential variable that can be instantiated for this
        parameter.

        Returns both the argument and the created variable.
        """
        from guppylang.tys.ty import ExistentialTypeVar

        var = ExistentialTypeVar.fresh(
            self.name, self.must_be_copyable, self.must_be_droppable
        )
        return TypeArg(var), var

    def to_bound(self, idx: int | None = None) -> TypeArg:
        """Creates a bound variable with a given index that can be instantiated for this
        parameter.
        """
        from guppylang.tys.ty import BoundTypeVar

        if idx is None:
            idx = self.idx
        return TypeArg(
            BoundTypeVar(self.name, idx, self.must_be_copyable, self.must_be_droppable)
        )

    def to_hugr(self) -> ht.TypeParam:
        """Computes the Hugr representation of the parameter."""
        return ht.TypeTypeParam(
            bound=ht.TypeBound.Any if self.can_be_linear else ht.TypeBound.Copyable
        )


@dataclass(frozen=True)
class ConstParam(ParameterBase):
    """A parameter of kind constant. Used to define fixed-size arrays etc."""

    ty: "Type"

    def __post_init__(self) -> None:
        if self.ty.unsolved_vars:
            raise InternalGuppyError(
                "Attempted to create constant param with unsolved type"
            )

    def with_idx(self, idx: int) -> "ConstParam":
        """Returns a copy of the parameter with a new index."""
        return ConstParam(idx, self.name, self.ty)

    def check_arg(self, arg: Argument, loc: AstNode | None = None) -> ConstArg:
        """Checks that this parameter can be instantiated with a given argument.

        Raises a user error if the argument is not valid.
        """
        match arg:
            case ConstArg(const):
                if const.ty != self.ty:
                    raise GuppyTypeError(
                        TypeMismatchError(loc, self.ty, const.ty, kind="argument")
                    )
                return arg
            case TypeArg(ty=ty):
                err = ExpectedError(
                    loc, f"expression of type `{self.ty}`", got=f"type `{ty}`"
                )
                raise GuppyTypeError(err)

    def to_existential(self) -> tuple[Argument, ExistentialVar]:
        """Creates a fresh existential variable that can be instantiated for this
        parameter.

        Returns both the argument and the created variable.
        """
        var = ExistentialConstVar.fresh(self.name, self.ty)
        return ConstArg(var), var

    def to_bound(self, idx: int | None = None) -> ConstArg:
        """Creates a bound variable with a given index that can be instantiated for this
        parameter.
        """
        if idx is None:
            idx = self.idx
        return ConstArg(BoundConstVar(self.ty, self.name, idx))

    def to_hugr(self) -> ht.TypeParam:
        """Computes the Hugr representation of the parameter."""
        from guppylang.tys.ty import NumericType

        match self.ty:
            case NumericType(kind=NumericType.Kind.Nat):
                return ht.BoundedNatParam(upper_bound=None)
            case _:
                assert isinstance(self.ty.to_hugr(), ht.ExtType)
                return ht.StringParam()


def check_all_args(
    params: Sequence[Parameter],
    args: Sequence[Argument],
    type_name: str,
    loc: AstNode | None = None,
) -> None:
    """Checks a list of arguments against the given parameters.

    Raises a user error if number of arguments doesn't match or one of the argument is
    invalid.
    """
    exp, act = len(params), len(args)
    if exp != act:
        # TODO: Adjust the error span to only point to the offending arguments (similar
        #  to how we deal with call args in the expression checker). This requires
        #  threading the type arg spans down to this point
        raise GuppyError(WrongNumberOfTypeArgsError(loc, exp, act, type_name))

    # Now check that the kinds match up
    for param, arg in zip(params, args, strict=True):
        # TODO: The error location is bad. We want the location of `arg`, not of the
        #  whole thing.
        param.check_arg(arg, loc)



"""Compilers building array functions on top of hugr standard operations."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Final, TypeVar

import hugr.std.collections
import hugr.std.int
import hugr.std.prelude
from hugr import Node, Wire, ops
from hugr import tys as ht
from hugr import val as hv

from guppylang.compiler.core import CompilerContext, GlobalConstId
from guppylang.definition.custom import CustomCallCompiler, CustomInoutCallCompiler
from guppylang.definition.value import CallReturnWires
from guppylang.error import InternalGuppyError
from guppylang.tys.subst import Inst

if TYPE_CHECKING:
    from collections.abc import Callable

    from hugr.build import function as hf
    from hugr.build.dfg import DfBase

    from guppylang.tys.subst import Inst


# --------------------------------------------
# --------------- prelude --------------------
# --------------------------------------------


def error_type() -> ht.ExtType:
    """Returns the hugr type of an error value."""
    return hugr.std.PRELUDE.types["error"].instantiate([])


@dataclass
class ErrorVal(hv.ExtensionValue):
    """Custom value for a floating point number."""

    signal: int
    message: str

    def to_value(self) -> hv.Extension:
        name = "ConstError"
        payload = {"signal": self.signal, "message": self.message}
        return hv.Extension(
            name, typ=error_type(), val=payload, extensions=[hugr.std.PRELUDE.name]
        )

    def __str__(self) -> str:
        return f"Error({self.signal}): {self.message}"


def panic(inputs: list[ht.Type], outputs: list[ht.Type]) -> ops.ExtOp:
    """Returns an operation that panics."""
    op_def = hugr.std.PRELUDE.get_op("panic")
    args: list[ht.TypeArg] = [
        ht.SequenceArg([ht.TypeTypeArg(ty) for ty in inputs]),
        ht.SequenceArg([ht.TypeTypeArg(ty) for ty in outputs]),
    ]
    sig = ht.FunctionType([error_type(), *inputs], outputs)
    return ops.ExtOp(op_def, sig, args)


# ------------------------------------------------------
# --------- Custom compilers for non-native ops --------
# ------------------------------------------------------


def build_panic(
    builder: DfBase[P],
    in_tys: ht.TypeRow,
    out_tys: ht.TypeRow,
    err: Wire,
    *args: Wire,
) -> Node:
    """Builds a panic operation."""
    op = panic(in_tys, out_tys)
    return builder.add_op(op, err, *args)


def build_error(builder: DfBase[P], signal: int, msg: str) -> Wire:
    """Constructs and loads a static error value."""
    val = ErrorVal(signal, msg)
    return builder.load(builder.add_const(val))


# TODO: Common up build_unwrap_right and build_unwrap_left below once
#  https://github.com/CQCL/hugr/issues/1596 is fixed


def build_unwrap_right(
    builder: DfBase[P], either: Wire, error_msg: str, error_signal: int = 1
) -> Node:
    """Unwraps the right value from a `hugr.tys.Either` value, panicking with the given
    message if the result is left.
    """
    conditional = builder.add_conditional(either)
    result_ty = builder.hugr.port_type(either.out_port())
    assert isinstance(result_ty, ht.Sum)
    [left_tys, right_tys] = result_ty.variant_rows
    with conditional.add_case(0) as case:
        error = build_error(case, error_signal, error_msg)
        case.set_outputs(*build_panic(case, left_tys, right_tys, error, *case.inputs()))
    with conditional.add_case(1) as case:
        case.set_outputs(*case.inputs())
    return conditional.to_node()


P = TypeVar("P", bound=ops.DfParentOp)


def build_unwrap_left(
    builder: DfBase[P], either: Wire, error_msg: str, error_signal: int = 1
) -> Node:
    """Unwraps the left value from a `hugr.tys.Either` value, panicking with the given
    message if the result is right.
    """
    conditional = builder.add_conditional(either)
    result_ty = builder.hugr.port_type(either.out_port())
    assert isinstance(result_ty, ht.Sum)
    [left_tys, right_tys] = result_ty.variant_rows
    with conditional.add_case(0) as case:
        case.set_outputs(*case.inputs())
    with conditional.add_case(1) as case:
        error = build_error(case, error_signal, error_msg)
        case.set_outputs(*build_panic(case, right_tys, left_tys, error, *case.inputs()))
    return conditional.to_node()


def build_unwrap(
    builder: DfBase[P], option: Wire, error_msg: str, error_signal: int = 1
) -> Node:
    """Unwraps an `hugr.tys.Option` value, panicking with the given message if the
    result is an error.
    """
    return build_unwrap_right(builder, option, error_msg, error_signal)


def build_expect_none(
    builder: DfBase[P], option: Wire, error_msg: str, error_signal: int = 1
) -> Node:
    """Checks that `hugr.tys.Option` value is `None`, otherwise panics with the given
    message.
    """
    return build_unwrap_left(builder, option, error_msg, error_signal)


class MemSwapCompiler(CustomCallCompiler):
    """Compiler for the `mem_swap` function."""

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [x, y] = args
        return CallReturnWires(regular_returns=[], inout_returns=[y, x])

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Call compile_with_inouts instead")


UNWRAP_RESULT: Final[GlobalConstId] = GlobalConstId.fresh("unwrap_result")


def _build_unwrap_result(func: hf.Function, result_type_var: ht.Variable) -> None:
    either = func.inputs()[0]
    conditional = func.add_conditional(either)
    with conditional.add_case(0) as case:
        [error] = list(case.inputs())
        case.set_outputs(
            *build_panic(case, [error_type()], [result_type_var], error, *case.inputs())
        )
    with conditional.add_case(1) as case:
        case.set_outputs(*case.inputs())
    func.set_outputs(*conditional.outputs())


def unwrap_result(
    builder: DfBase[P],
    ctx: CompilerContext,
    either: Wire,
) -> Wire:
    """Builds or retrieves and then calls a function that unwraps an `hugr.tys.Either`
    value, panicking if the result is an error.
    """
    either_ty = builder.hugr.port_type(either.out_port())
    assert isinstance(either_ty, ht.Either)
    [error_tys, result_tys] = either_ty.variant_rows
    # Construct the function signature for unwrapping a result of type T.
    func_ty = ht.PolyFuncType(
        params=[ht.TypeTypeParam(ht.TypeBound.Any)],
        body=ht.FunctionType(
            input=[ht.Either(error_tys, [ht.Variable(0, ht.TypeBound.Any)])],
            output=[ht.Variable(0, ht.TypeBound.Any)],
        ),
    )
    # Build global unwrap result function if it doesn't already exist.
    func, already_exists = ctx.declare_global_func(UNWRAP_RESULT, func_ty)
    if not already_exists:
        _build_unwrap_result(func, ht.Variable(0, ht.TypeBound.Any))
    # Call the global function.
    concrete_ty = ht.FunctionType(
        input=[ht.Either(error_tys, result_tys)], output=result_tys
    )
    type_args = [ht.TypeTypeArg(*result_tys)]
    func_call = builder.call(
        func.parent_node,
        either,
        instantiation=concrete_ty,
        type_args=type_args,
    )
    [result] = list(func_call.outputs())
    return result


class UnwrapOpCompiler(CustomInoutCallCompiler):
    """Compiler for operations that require unwrapping a result which could potentially
    cause a panic.

    Args:
        op: A HUGR operation that outputs an Either<error, result> value.
    """

    op: Callable[[ht.FunctionType, Inst], ops.DataflowOp]

    def __init__(self, op: Callable[[ht.FunctionType, Inst], ops.DataflowOp]):
        self.op = op

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        assert len(self.ty.output) == 1
        # To instantiate the op we need a function signature that wraps the output of
        # the function that is being compiled into an either type.
        opt_func_type = ht.FunctionType(
            input=self.ty.input,
            output=[ht.Either([error_type()], self.ty.output)],
        )
        op = self.op(opt_func_type, self.type_args)
        either = self.builder.add_op(op, *args)
        result = unwrap_result(self.builder, self.ctx, either)
        return CallReturnWires(regular_returns=[result], inout_returns=[])


class BarrierCompiler(CustomCallCompiler):
    """Compiler for the `barrier` function."""

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        tys = [t for arg in args if (t := self.builder.hugr.port_type(arg.out_port()))]

        op = hugr.std.prelude.PRELUDE_EXTENSION.get_op("Barrier").instantiate(
            [ht.SequenceArg([ht.TypeTypeArg(ty) for ty in tys])]
        )

        barrier_n = self.builder.add_op(op, *args)

        return CallReturnWires(
            regular_returns=[], inout_returns=[barrier_n[i] for i in range(len(tys))]
        )

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Call compile_with_inouts instead")



import argparse
from pathlib import Path

import guppylang

guppylang.enable_experimental_features()


def pytest_addoption(parser):
    def dir_path(s):
        path = Path(s)
        if not path.exists() or path.is_dir():
            return path
        msg = f"export-test-cases dir:{path} exists and is not a directory"
        raise argparse.ArgumentTypeError(msg)

    parser.addoption(
        "--export-test-cases",
        action="store",
        type=dir_path,
        help="A directory to which to export test cases",
    )

    parser.addoption(
        "--no_validation",
        dest="validation",
        action="store_false",
        help="Disable validation tests (run by default)",
    )



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load_all(quantum)

@guppy.struct(module)
class MyStruct:
    q: qubit


@guppy(module)
def foo(s: MyStruct @owned) -> tuple[qubit, MyStruct]:
    return s.q, s


module.compile()



"""Guppy module for builtin types and operations."""

# mypy: disable-error-code="empty-body, misc, override, valid-type, no-untyped-def"

from typing import Any, Generic, TypeVar, no_type_check

import hugr.std.int

from guppylang.decorator import guppy
from guppylang.definition.custom import CopyInoutCompiler, NoopCompiler
from guppylang.std._internal.checker import (
    ArrayCopyChecker,
    BarrierChecker,
    CallableChecker,
    DunderChecker,
    NewArrayChecker,
    PanicChecker,
    RangeChecker,
    ResultChecker,
    ReversingChecker,
    UnsupportedChecker,
)
from guppylang.std._internal.compiler.array import (
    ArrayGetitemCompiler,
    ArrayIterAsertAllUsedCompiler,
    ArraySetitemCompiler,
    NewArrayCompiler,
)
from guppylang.std._internal.compiler.list import (
    ListGetitemCompiler,
    ListLengthCompiler,
    ListPopCompiler,
    ListPushCompiler,
    ListSetitemCompiler,
)
from guppylang.std._internal.compiler.prelude import (
    MemSwapCompiler,
    UnwrapOpCompiler,
)
from guppylang.std._internal.util import (
    float_op,
    int_op,
    logic_op,
    unsupported_op,
)
from guppylang.tys.builtin import (
    array_type_def,
    bool_type_def,
    float_type_def,
    int_type_def,
    list_type_def,
    nat_type_def,
    sized_iter_type_def,
    string_type_def,
)

guppy.init_module(import_builtins=False)

T = guppy.type_var("T")
L = guppy.type_var("L", copyable=False, droppable=False)


def py(*args: Any) -> Any:
    """Function to tag compile-time evaluated Python expressions in a Guppy context.

    This function acts like the identity when execute in a Python context.
    """
    return tuple(args)


class _Owned:
    """Dummy class to support `@owned` annotations."""

    def __rmatmul__(self, other: Any) -> Any:
        return other


owned = _Owned()


class nat:
    """Class to import in order to use nats."""


_T = TypeVar("_T")
_n = TypeVar("_n")


class array(Generic[_T, _n]):
    """Class to import in order to use arrays."""

    def __init__(self, *args: Any):
        pass


@guppy.extend_type(bool_type_def)
class Bool:
    @guppy.hugr_op(logic_op("And"))
    def __and__(self: bool, other: bool) -> bool: ...

    @guppy.custom(NoopCompiler())
    def __bool__(self: bool) -> bool: ...

    @guppy.hugr_op(logic_op("Eq"))
    def __eq__(self: bool, other: bool) -> bool: ...

    @guppy
    @no_type_check
    def __ne__(self: bool, other: bool) -> bool:
        return not self == other

    @guppy
    @no_type_check
    def __int__(self: bool) -> int:
        return 1 if self else 0

    @guppy
    @no_type_check
    def __nat__(self: bool) -> nat:
        # TODO: Type information doesn't flow through the `if` expression, so we
        #  have to insert the `nat` coercions by hand.
        #  See https://github.com/CQCL/guppylang/issues/707
        return nat(1) if self else nat(0)

    @guppy.custom(checker=DunderChecker("__bool__"), higher_order_value=False)
    def __new__(x): ...

    @guppy.hugr_op(logic_op("Or"))
    def __or__(self: bool, other: bool) -> bool: ...

    @guppy.hugr_op(logic_op("Xor"))
    def __xor__(self: bool, other: bool) -> bool: ...


@guppy.extend_type(string_type_def)
class String:
    @guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
    def __new__(x): ...


@guppy.extend_type(nat_type_def)
class Nat:
    @guppy.custom(NoopCompiler())
    def __abs__(self: nat) -> nat: ...

    @guppy.hugr_op(int_op("iadd"))
    def __add__(self: nat, other: nat) -> nat: ...

    @guppy.hugr_op(int_op("iand"))
    def __and__(self: nat, other: nat) -> nat: ...

    @guppy
    @no_type_check
    def __bool__(self: nat) -> bool:
        return self != 0

    @guppy.custom(NoopCompiler())
    def __ceil__(self: nat) -> nat: ...

    @guppy.hugr_op(int_op("idivmod_u", n_vars=2))
    def __divmod__(self: nat, other: nat) -> tuple[nat, nat]: ...

    @guppy.hugr_op(int_op("ieq"))
    def __eq__(self: nat, other: nat) -> bool: ...

    @guppy.hugr_op(int_op("convert_u", hugr.std.int.CONVERSIONS_EXTENSION))
    def __float__(self: nat) -> float: ...

    @guppy.custom(NoopCompiler())
    def __floor__(self: nat) -> nat: ...

    @guppy.hugr_op(int_op("idiv_u"))
    def __floordiv__(self: nat, other: nat) -> nat: ...

    @guppy.hugr_op(int_op("ige_u"))
    def __ge__(self: nat, other: nat) -> bool: ...

    @guppy.hugr_op(int_op("igt_u"))
    def __gt__(self: nat, other: nat) -> bool: ...

    # TODO: Use "iu_to_s" once we have lowering:
    #  https://github.com/CQCL/hugr/issues/1806
    @guppy.custom(NoopCompiler())
    def __int__(self: nat) -> int: ...

    @guppy.hugr_op(int_op("inot"))
    def __invert__(self: nat) -> nat: ...

    @guppy.hugr_op(int_op("ile_u"))
    def __le__(self: nat, other: nat) -> bool: ...

    @guppy.hugr_op(int_op("ishl", n_vars=2))
    def __lshift__(self: nat, other: nat) -> nat: ...

    @guppy.hugr_op(int_op("ilt_u"))
    def __lt__(self: nat, other: nat) -> bool: ...

    @guppy.hugr_op(int_op("imod_u", n_vars=2))
    def __mod__(self: nat, other: nat) -> int: ...

    @guppy.hugr_op(int_op("imul"))
    def __mul__(self: nat, other: nat) -> nat: ...

    @guppy.custom(NoopCompiler())
    def __nat__(self: nat) -> nat: ...

    @guppy.hugr_op(int_op("ine"))
    def __ne__(self: nat, other: nat) -> bool: ...

    @guppy.custom(checker=DunderChecker("__nat__"), higher_order_value=False)
    def __new__(x): ...

    @guppy.hugr_op(int_op("ior"))
    def __or__(self: nat, other: nat) -> nat: ...

    @guppy.custom(NoopCompiler())
    def __pos__(self: nat) -> nat: ...

    @guppy.hugr_op(int_op("ipow"))
    def __pow__(self: nat, other: nat) -> nat: ...

    @guppy.custom(checker=ReversingChecker())
    def __radd__(self: nat, other: nat) -> nat: ...

    @guppy.custom(checker=ReversingChecker())
    def __rand__(self: nat, other: nat) -> nat: ...

    @guppy.custom(checker=ReversingChecker())
    def __rdivmod__(self: nat, other: nat) -> tuple[nat, nat]: ...

    @guppy.custom(checker=ReversingChecker())
    def __rfloordiv__(self: nat, other: nat) -> nat: ...

    @guppy.custom(checker=ReversingChecker())
    def __rlshift__(self: nat, other: nat) -> nat: ...

    @guppy.custom(checker=ReversingChecker())
    def __rmod__(self: nat, other: nat) -> nat: ...

    @guppy.custom(checker=ReversingChecker())
    def __rmul__(self: nat, other: nat) -> nat: ...

    @guppy.custom(checker=ReversingChecker())
    def __ror__(self: nat, other: nat) -> nat: ...

    @guppy.custom(NoopCompiler())
    def __round__(self: nat) -> nat: ...

    @guppy.custom(checker=ReversingChecker())
    def __rpow__(self: nat, other: nat) -> nat: ...

    @guppy.custom(checker=ReversingChecker())
    def __rrshift__(self: nat, other: nat) -> nat: ...

    @guppy.hugr_op(int_op("ishr"))
    def __rshift__(self: nat, other: nat) -> nat: ...

    @guppy.custom(checker=ReversingChecker())
    def __rsub__(self: nat, other: nat) -> nat: ...

    @guppy.custom(checker=ReversingChecker())
    def __rtruediv__(self: nat, other: nat) -> float: ...

    @guppy.custom(checker=ReversingChecker())
    def __rxor__(self: nat, other: nat) -> nat: ...

    @guppy.hugr_op(int_op("isub"))
    def __sub__(self: nat, other: nat) -> nat: ...

    @guppy
    @no_type_check
    def __truediv__(self: nat, other: nat) -> float:
        return float(self) / float(other)

    @guppy.custom(NoopCompiler())
    def __trunc__(self: nat) -> nat: ...

    @guppy.hugr_op(int_op("ixor"))
    def __xor__(self: nat, other: nat) -> nat: ...


@guppy.extend_type(int_type_def)
class Int:
    @guppy.hugr_op(int_op("iabs"))  # TODO: Maybe wrong? (signed vs unsigned!)
    def __abs__(self: int) -> int: ...

    @guppy.hugr_op(int_op("iadd"))
    def __add__(self: int, other: int) -> int: ...

    @guppy.hugr_op(int_op("iand"))
    def __and__(self: int, other: int) -> int: ...

    @guppy
    @no_type_check
    def __bool__(self: int) -> bool:
        return self != 0

    @guppy.custom(NoopCompiler())
    def __ceil__(self: int) -> int: ...

    @guppy.hugr_op(int_op("idivmod_s"))
    def __divmod__(self: int, other: int) -> tuple[int, int]: ...

    @guppy.hugr_op(int_op("ieq"))
    def __eq__(self: int, other: int) -> bool: ...

    @guppy.hugr_op(int_op("convert_s", hugr.std.int.CONVERSIONS_EXTENSION))
    def __float__(self: int) -> float: ...

    @guppy.custom(NoopCompiler())
    def __floor__(self: int) -> int: ...

    @guppy.hugr_op(int_op("idiv_s"))
    def __floordiv__(self: int, other: int) -> int: ...

    @guppy.hugr_op(int_op("ige_s"))
    def __ge__(self: int, other: int) -> bool: ...

    @guppy.hugr_op(int_op("igt_s"))
    def __gt__(self: int, other: int) -> bool: ...

    @guppy.custom(NoopCompiler())
    def __int__(self: int) -> int: ...

    @guppy.hugr_op(int_op("inot"))
    def __invert__(self: int) -> int: ...

    @guppy.hugr_op(int_op("ile_s"))
    def __le__(self: int, other: int) -> bool: ...

    @guppy.hugr_op(int_op("ishl"))  # TODO: RHS is unsigned
    def __lshift__(self: int, other: int) -> int: ...

    @guppy.hugr_op(int_op("ilt_s"))
    def __lt__(self: int, other: int) -> bool: ...

    @guppy.hugr_op(int_op("imod_s"))
    def __mod__(self: int, other: int) -> int: ...

    @guppy.hugr_op(int_op("imul"))
    def __mul__(self: int, other: int) -> int: ...

    @guppy.hugr_op(int_op("is_to_u"))  # TODO
    def __nat__(self: int) -> nat: ...

    @guppy.hugr_op(int_op("ine"))
    def __ne__(self: int, other: int) -> bool: ...

    @guppy.hugr_op(int_op("ineg"))
    def __neg__(self: int) -> int: ...

    @guppy.custom(checker=DunderChecker("__int__"), higher_order_value=False)
    def __new__(x): ...

    @guppy.hugr_op(int_op("ior"))
    def __or__(self: int, other: int) -> int: ...

    @guppy.custom(NoopCompiler())
    def __pos__(self: int) -> int: ...

    @guppy
    @no_type_check
    def __pow__(self: int, exponent: int) -> int:
        if exponent < 0:
            panic(
                "Negative exponent not supported in"
                "__pow__ with int type base. Try casting the base to float."
            )
        return self.__pow_impl(exponent)

    @guppy.hugr_op(int_op("ipow"))
    def __pow_impl(self: int, exponent: int) -> int: ...

    @guppy.custom(checker=ReversingChecker())
    def __radd__(self: int, other: int) -> int: ...

    @guppy.custom(checker=ReversingChecker())
    def __rand__(self: int, other: int) -> int: ...

    @guppy.custom(checker=ReversingChecker())
    def __rdivmod__(self: int, other: int) -> tuple[int, int]: ...

    @guppy.custom(checker=ReversingChecker())
    def __rfloordiv__(self: int, other: int) -> int: ...

    @guppy.custom(checker=ReversingChecker())  # TODO: RHS is unsigned
    def __rlshift__(self: int, other: int) -> int: ...

    @guppy.custom(checker=ReversingChecker())
    def __rmod__(self: int, other: int) -> int: ...

    @guppy.custom(checker=ReversingChecker())
    def __rmul__(self: int, other: int) -> int: ...

    @guppy.custom(checker=ReversingChecker())
    def __ror__(self: int, other: int) -> int: ...

    @guppy.custom(NoopCompiler())
    def __round__(self: int) -> int: ...

    @guppy.custom(checker=ReversingChecker())
    def __rpow__(self: int, other: int) -> int: ...

    @guppy.custom(checker=ReversingChecker())  # TODO: RHS is unsigned
    def __rrshift__(self: int, other: int) -> int: ...

    @guppy.custom(checker=ReversingChecker())  # TODO: RHS is unsigned
    def __rshift__(self: int, other: int) -> int: ...

    @guppy.custom(checker=ReversingChecker())
    def __rsub__(self: int, other: int) -> int: ...

    @guppy.custom(checker=ReversingChecker())
    def __rtruediv__(self: int, other: int) -> float: ...

    @guppy.custom(checker=ReversingChecker())
    def __rxor__(self: int, other: int) -> int: ...

    @guppy.hugr_op(int_op("isub"))
    def __sub__(self: int, other: int) -> int: ...

    @guppy
    @no_type_check
    def __truediv__(self: int, other: int) -> float:
        return float(self) / float(other)

    @guppy.custom(NoopCompiler())
    def __trunc__(self: int) -> int: ...

    @guppy.hugr_op(int_op("ixor"))
    def __xor__(self: int, other: int) -> int: ...


@guppy.extend_type(float_type_def)
class Float:
    @guppy.hugr_op(float_op("fabs"))
    def __abs__(self: float) -> float: ...

    @guppy.hugr_op(float_op("fadd"))
    def __add__(self: float, other: float) -> float: ...

    @guppy
    @no_type_check
    def __bool__(self: float) -> bool:
        return self != 0.0

    @guppy.hugr_op(float_op("fceil"))
    def __ceil__(self: float) -> float: ...

    @guppy
    @no_type_check
    def __divmod__(self: float, other: float) -> tuple[float, float]:
        return self // other, self.__mod__(other)

    @guppy.hugr_op(float_op("feq"))
    def __eq__(self: float, other: float) -> bool: ...

    @guppy.custom(NoopCompiler())
    def __float__(self: float) -> float: ...

    @guppy.hugr_op(float_op("ffloor"))
    def __floor__(self: float) -> float: ...

    @guppy
    @no_type_check
    def __floordiv__(self: float, other: float) -> float:
        return (self / other).__floor__()

    @guppy.hugr_op(float_op("fge"))
    def __ge__(self: float, other: float) -> bool: ...

    @guppy.hugr_op(float_op("fgt"))
    def __gt__(self: float, other: float) -> bool: ...

    @guppy.custom(
        UnwrapOpCompiler(
            # Use `int_op` to instantiate type arg with 64-bit integer.
            int_op("trunc_s", hugr.std.int.CONVERSIONS_EXTENSION),
        )
    )
    def __int__(self: float) -> int: ...

    @guppy.hugr_op(float_op("fle"))
    def __le__(self: float, other: float) -> bool: ...

    @guppy.hugr_op(float_op("flt"))
    def __lt__(self: float, other: float) -> bool: ...

    @guppy
    @no_type_check
    def __mod__(self: float, other: float) -> float:
        return self - (self // other) * other

    @guppy.hugr_op(float_op("fmul"))
    def __mul__(self: float, other: float) -> float: ...

    @guppy.custom(
        UnwrapOpCompiler(
            # Use `int_op` to instantiate type arg with 64-bit integer.
            int_op("trunc_u", hugr.std.int.CONVERSIONS_EXTENSION),
        )
    )
    def __nat__(self: float) -> nat: ...

    @guppy.hugr_op(float_op("fne"))
    def __ne__(self: float, other: float) -> bool: ...

    @guppy.hugr_op(float_op("fneg"))
    def __neg__(self: float) -> float: ...

    @guppy.custom(checker=DunderChecker("__float__"), higher_order_value=False)
    def __new__(x): ...

    @guppy.custom(NoopCompiler())
    def __pos__(self: float) -> float: ...

    @guppy.hugr_op(float_op("fpow"))  # TODO
    def __pow__(self: float, other: float) -> float: ...

    @guppy.custom(checker=ReversingChecker())
    def __radd__(self: float, other: float) -> float: ...

    @guppy.custom(checker=ReversingChecker())
    def __rdivmod__(self: float, other: float) -> tuple[float, float]: ...

    @guppy.custom(checker=ReversingChecker())
    def __rfloordiv__(self: float, other: float) -> float: ...

    @guppy.custom(checker=ReversingChecker())
    def __rmod__(self: float, other: float) -> float: ...

    @guppy.custom(checker=ReversingChecker())
    def __rmul__(self: float, other: float) -> float: ...

    @guppy.hugr_op(float_op("fround"))  # TODO
    def __round__(self: float) -> float: ...

    @guppy.custom(checker=ReversingChecker())
    def __rpow__(self: float, other: float) -> float: ...

    @guppy.custom(checker=ReversingChecker())
    def __rsub__(self: float, other: float) -> float: ...

    @guppy.custom(checker=ReversingChecker())
    def __rtruediv__(self: float, other: float) -> float: ...

    @guppy.hugr_op(float_op("fsub"))
    def __sub__(self: float, other: float) -> float: ...

    @guppy.hugr_op(float_op("fdiv"))
    def __truediv__(self: float, other: float) -> float: ...

    @guppy.hugr_op(unsupported_op("trunc_s"))  # TODO `trunc_s` returns an option
    def __trunc__(self: float) -> float: ...


@guppy.extend_type(list_type_def)
class List:
    @guppy.custom(ListGetitemCompiler())
    def __getitem__(self: list[L], idx: int) -> L: ...

    @guppy.custom(ListSetitemCompiler())
    def __setitem__(self: list[L], idx: int, value: L @ owned) -> None: ...

    @guppy.custom(ListLengthCompiler())
    def __len__(self: list[L]) -> int: ...

    @guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
    def __new__(x): ...

    @guppy.custom(NoopCompiler())  # TODO: define via Guppy source instead
    def __iter__(self: list[L] @ owned) -> list[L]: ...

    @guppy.hugr_op(unsupported_op("pop"))
    def __next__(self: list[L] @ owned) -> "Option[tuple[L, list[L]]]": ...  # type: ignore[type-arg]

    @guppy.custom(ListPushCompiler())
    def append(self: list[L], item: L @ owned) -> None: ...

    @guppy.custom(ListPopCompiler())
    def pop(self: list[L]) -> L: ...


n = guppy.nat_var("n")


@guppy.extend_type(array_type_def)
class Array:
    @guppy.custom(ArrayGetitemCompiler())
    def __getitem__(self: array[L, n], idx: int) -> L: ...

    @guppy.custom(ArraySetitemCompiler())
    def __setitem__(self: array[L, n], idx: int, value: L @ owned) -> None: ...

    @guppy
    @no_type_check
    def __len__(self: array[L, n]) -> int:
        return n

    @guppy.custom(NewArrayCompiler(), NewArrayChecker(), higher_order_value=False)
    def __new__(): ...

    @guppy
    @no_type_check
    def __iter__(self: array[L, n] @ owned) -> "SizedIter[ArrayIter[L, n], n]":
        return SizedIter(ArrayIter(self, 0))

    @guppy.custom(CopyInoutCompiler(), ArrayCopyChecker())
    def copy(self: array[T, n]) -> array[T, n]: ...


@guppy.struct
class ArrayIter(Generic[L, n]):
    xs: array[L, n]
    i: int

    @guppy
    @no_type_check
    def __next__(
        self: "ArrayIter[L, n]" @ owned,
    ) -> "Option[tuple[L, ArrayIter[L, n]]]":
        if self.i < int(n):
            elem = _array_unsafe_getitem(self.xs, self.i)
            return some((elem, ArrayIter(self.xs, self.i + 1)))
        self._assert_all_used()
        return nothing()

    @guppy.custom(ArrayIterAsertAllUsedCompiler())
    def _assert_all_used(self: "ArrayIter[L, n]" @ owned) -> None: ...


@guppy.custom(ArrayGetitemCompiler())
def _array_unsafe_getitem(xs: array[L, n], idx: int) -> L: ...


@guppy.extend_type(sized_iter_type_def)
class SizedIter:
    """A wrapper around an iterator type `L` promising that the iterator will yield
    exactly `n` values.

    Annotating an iterator with an incorrect size is undefined behaviour.
    """

    def __class_getitem__(cls, item: Any) -> type:
        # Dummy implementation to allow subscripting of the `SizedIter` type in
        # positions that are evaluated by the Python interpreter
        return cls

    @guppy.custom(NoopCompiler())
    def __new__(iterator: L @ owned) -> "SizedIter[L, n]":  # type: ignore[type-arg]
        """Casts an iterator into a `SizedIter`."""

    @guppy.custom(NoopCompiler())
    def unwrap_iter(self: "SizedIter[L, n]" @ owned) -> L:
        """Extracts the actual iterator."""

    @guppy.custom(NoopCompiler())
    def __iter__(self: "SizedIter[L, n]" @ owned) -> "SizedIter[L, n]":  # type: ignore[type-arg]
        """Dummy implementation making sized iterators iterable themselves."""


# TODO: This is a temporary hack until we have implemented the proper results mechanism.
@guppy.custom(checker=ResultChecker(), higher_order_value=False)
def result(tag, value): ...


@guppy.custom(checker=PanicChecker(), higher_order_value=False)
def panic(msg, *args):
    """Panic, throwing an error with the given message, and immediately exit the
    program.

    Return type is arbitrary, as this function never returns.

    Args:
        msg: The message to display. Must be a string literal.
        args: Arbitrary extra inputs, will not affect the message. Only useful for
        consuming linear values.
    """


@guppy.custom(checker=DunderChecker("__abs__"), higher_order_value=False)
def abs(x): ...


@guppy.custom(checker=CallableChecker(), higher_order_value=False)
def callable(x): ...


@guppy.custom(checker=DunderChecker("__divmod__", num_args=2), higher_order_value=False)
def divmod(x, y): ...


@guppy.custom(checker=DunderChecker("__len__"), higher_order_value=False)
def len(x): ...


@guppy.custom(checker=DunderChecker("__pow__", num_args=2), higher_order_value=False)
def pow(x, y): ...


@guppy.custom(checker=DunderChecker("__round__"), higher_order_value=False)
def round(x): ...


# Python builtins that are not supported yet:


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def aiter(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def all(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def anext(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def any(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def bin(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def breakpoint(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def bytearray(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def bytes(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def chr(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def classmethod(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def compile(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def complex(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def delattr(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def dict(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def dir(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def enumerate(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def eval(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def exec(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def filter(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def format(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def forozenset(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def getattr(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def globals(): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def hasattr(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def hash(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def help(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def hex(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def id(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def input(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def isinstance(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def issubclass(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def iter(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def locals(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def map(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def max(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def memoryview(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def min(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def next(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def object(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def oct(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def open(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def ord(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def print(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def property(x): ...


@guppy.struct
class Range:
    next: int
    stop: int

    @guppy
    def __iter__(self: "Range") -> "Range":
        return self

    @guppy
    @no_type_check
    def __next__(self: "Range") -> "Option[tuple[int, Range]]":
        if self.next < self.stop:
            return some((self.next, Range(self.next + 1, self.stop)))
        return nothing()


@guppy.custom(checker=RangeChecker(), higher_order_value=False)
def range(stop: int) -> Range:
    """Limited version of python range().
    Only a single argument (stop/limit) is supported."""


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def repr(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def reversed(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def set(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def setattr(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def slice(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def sorted(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def staticmethod(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def sum(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def super(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def type(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def vars(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def zip(x): ...


@guppy.custom(checker=UnsupportedChecker(), higher_order_value=False)
def __import__(x): ...


@guppy.custom(MemSwapCompiler())
def mem_swap(x: L, y: L) -> None:
    """Swaps the values of two variables."""


@guppy.custom(checker=BarrierChecker(), higher_order_value=False)
def barrier(*args) -> None:
    """Barrier to guarantee that all operations before the barrier are completed before
    operations after the barrier are started."""


# Import `Option` since it's part of the default module. Has to be at the end of the
# file to avoid cyclic import.
#
# TODO: This is a temporary solution until https://github.com/CQCL/guppylang/issues/732
#  is properly addressed.
from guppylang.std.option import Option, nothing, some  # noqa: E402



from typing import Generic

from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")
T = guppy.type_var("T", module=module)


@guppy.struct(module)
class MyStruct(Generic[T]):
    x: T
    y: T


@guppy(module)
def main() -> None:
    MyStruct(0, False)


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array


module = GuppyModule("test")

@guppy(module)
def main() -> None:
    a = array(1, 2, 3)
    a[0] = "not an int"

module.compile()


import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load_all(quantum)


@guppy.declare(module)
def new_qubit() -> qubit:
    ...


@guppy(module)
def foo(b: bool) -> qubit:
    if b:
        q = new_qubit()
    else:
        q = new_qubit()
    q = new_qubit()
    return q


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit

module = GuppyModule("test")
module.load(qubit)


@guppy.declare(module)
def foo(q: qubit) -> None: ...


@guppy(module)
def test(q: qubit @owned) -> None:
   foo(q)


module.compile()



"""Tests for using python expressions in guppy functions."""

from importlib.util import find_spec


from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import py, array, nat, owned
from tests.util import compile_guppy

tket2_installed = find_spec("tket2") is not None


def test_basic(validate):
    x = 42

    @compile_guppy
    def foo() -> int:
        return py(x + 1)

    validate(foo)


def test_builtin(validate):
    @compile_guppy
    def foo() -> int:
        return py(len({"a": 1337, "b": None}))

    validate(foo)


def test_if(validate):
    b = True

    @compile_guppy
    def foo() -> int:
        if py(b or 1 > 6):
            return 0
        return 1

    validate(foo)


def test_redeclare_after(validate):
    x = 1

    @compile_guppy
    def foo() -> int:
        return py(x)

    x = False

    validate(foo)


def test_tuple(validate):
    @compile_guppy
    def foo() -> int:
        x, y = py((1, False))
        return x

    validate(foo)


def test_tuple_implicit(validate):
    @compile_guppy
    def foo() -> int:
        x, y = py(1, False)
        return x

    validate(foo)


def test_list_basic(validate):
    @compile_guppy
    def foo() -> array[int, 3]:
        xs = py([1, 2, 3])
        return xs

    validate(foo)


def test_list_empty(validate):
    @compile_guppy
    def foo() -> array[int, 0]:
        return py([])

    validate(foo)


def test_list_empty_nested(validate):
    @compile_guppy
    def foo() -> None:
        xs: array[tuple[int, array[bool, 0]], 1] = py([(42, [])])

    validate(foo)


def test_list_empty_multiple(validate):
    @compile_guppy
    def foo() -> None:
        xs: tuple[array[int, 0], array[bool, 0]] = py([], [])

    validate(foo)


def test_nats_from_ints(validate):
    @compile_guppy
    def foo() -> None:
        x: nat = py(1)
        y: tuple[nat, nat] = py(2, 3)
        z: array[nat, 3] = py([4, 5, 6])

    validate(foo)


def test_strings(validate):
    @compile_guppy
    def foo() -> None:
        x: str = py("a" + "b")

    validate(foo)


def test_func_type_arg(validate):
    module = GuppyModule("test")
    n = 10

    @guppy(module)
    def foo(xs: array[int, py(n)] @ owned) -> array[int, py(n)]:
        return xs

    @guppy.declare(module)
    def bar(xs: array[int, py(n)]) -> array[int, py(n)]: ...

    @guppy.struct(module)
    class Baz:
        xs: array[int, py(n)]

    validate(module.compile())



from typing import no_type_check

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std._internal.compiler.quantum import QSYSTEM_UTILS_EXTENSION
from guppylang.std._internal.util import external_op

qsystem_utils = GuppyModule("qsystem.utils")


@guppy.hugr_op(
    external_op("GetCurrentShot", [], ext=QSYSTEM_UTILS_EXTENSION), module=qsystem_utils
)
@no_type_check
def get_current_shot() -> int: ...



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit, measure


module = GuppyModule("test")
module.load_all(quantum)


@guppy.struct(module)
class MyStruct:
    q: qubit


@guppy(module)
def foo(b: bool) -> bool:
    s = MyStruct(qubit())
    if b:
        return measure(s.q)
    return False


module.compile()



"""Tests for loading pytket circuits as functions."""

from importlib.util import find_spec

import pytest

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std import quantum
from guppylang.std.quantum import qubit, discard_array
from guppylang.std.builtins import array

tket2_installed = find_spec("tket2") is not None


@pytest.mark.skipif(not tket2_installed, reason="Tket2 is not installed")
def test_single_qubit_circuit(validate):
    from pytket import Circuit

    circ = Circuit(1)
    circ.H(0)

    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.pytket(circ, module)
    def guppy_circ(q1: qubit) -> None: ...

    @guppy(module)
    def foo(q: qubit) -> None:
        guppy_circ(q)

    validate(module.compile())


@pytest.mark.skipif(not tket2_installed, reason="Tket2 is not installed")
def test_multi_qubit_circuit(validate):
    from pytket import Circuit

    circ = Circuit(2)
    circ.H(0)
    circ.CX(0, 1)

    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.pytket(circ, module)
    def guppy_circ(q1: qubit, q2: qubit) -> None: ...

    @guppy(module)
    def foo(q1: qubit, q2: qubit) -> None:
        guppy_circ(q1, q2)

    validate(module.compile())


@pytest.mark.skipif(not tket2_installed, reason="Tket2 is not installed")
def test_measure(validate):
    from pytket import Circuit

    circ = Circuit(1)
    circ.H(0)
    circ.measure_all()

    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.pytket(circ, module)
    def guppy_circ(q: qubit) -> bool: ...

    @guppy(module)
    def foo(q: qubit) -> bool:
        return guppy_circ(q)

    validate(module.compile())


@pytest.mark.skipif(not tket2_installed, reason="Tket2 is not installed")
def test_measure_multiple(validate):
    from pytket import Circuit

    circ = Circuit(2, 2)
    circ.H(0)
    circ.measure_all()

    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.pytket(circ, module)
    def guppy_circ(q1: qubit, q2: qubit) -> tuple[bool, bool]: ...

    @guppy(module)
    def foo(q1: qubit, q2: qubit) -> tuple[bool, bool]:
        return guppy_circ(q1, q2)

    validate(module.compile())


@pytest.mark.skipif(not tket2_installed, reason="Tket2 is not installed")
def test_measure_not_last(validate):
    from pytket import Circuit

    circ = Circuit(1, 1)
    circ.H(0)
    circ.measure_all()
    circ.X(0)

    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.pytket(circ, module)
    def guppy_circ(q: qubit) -> bool: ...

    @guppy(module)
    def foo(q: qubit) -> bool:
        return guppy_circ(q)

    validate(module.compile())


@pytest.mark.skipif(not tket2_installed, reason="Tket2 is not installed")
def test_load_circuit(validate):
    from pytket import Circuit

    circ = Circuit(1)
    circ.H(0)

    module = GuppyModule("test")
    module.load_all(quantum)

    guppy.load_pytket("guppy_circ", circ, module, use_arrays=False)

    @guppy(module)
    def foo(q: qubit) -> None:
        guppy_circ(q)

    validate(module.compile())


@pytest.mark.skipif(not tket2_installed, reason="Tket2 is not installed")
def test_load_circuits(validate):
    from pytket import Circuit

    circ1 = Circuit(1)
    circ1.H(0)

    circ2 = Circuit(2)
    circ2.CX(0, 1)
    circ2.measure_all()

    module = GuppyModule("test")
    module.load_all(quantum)

    guppy.load_pytket("guppy_circ1", circ1, module, use_arrays=False)
    guppy.load_pytket("guppy_circ2", circ2, module, use_arrays=False)

    @guppy(module)
    def foo(q1: qubit, q2: qubit, q3: qubit) -> tuple[bool, bool]:
        guppy_circ1(q1)
        return  guppy_circ2(q2, q3)

    validate(module.compile())


@pytest.mark.skipif(not tket2_installed, reason="Tket2 is not installed")
def test_measure_some(validate):
    from pytket import Circuit

    circ = Circuit(2, 1)
    circ.CX(0, 1)
    circ.Measure(0, 0)

    module = GuppyModule("test")
    module.load_all(quantum)

    guppy.load_pytket("guppy_circ", circ, module, use_arrays=False)

    @guppy(module)
    def foo(q1: qubit, q2: qubit) -> bool:
        return  guppy_circ(q1, q2)

    validate(module.compile())


@pytest.mark.skipif(not tket2_installed, reason="Tket2 is not installed")
def test_register_arrays_default(validate):
    from pytket import Circuit

    circ = Circuit(2)

    module = GuppyModule("test")
    module.load_all(quantum)

    guppy.load_pytket("guppy_circ", circ, module)

    @guppy(module)
    def foo(default_reg: array[qubit, 2]) -> None:
        return guppy_circ(default_reg)

    # print(module.compile_hugr().render_dot())
    validate(module.compile())


@pytest.mark.skipif(not tket2_installed, reason="Tket2 is not installed")
def test_register_arrays(validate):
    from pytket import Circuit

    circ = Circuit(2)
    reg = circ.add_q_register("extra_reg", 3)
    circ.measure_register(reg, "extra_bits")

    module = GuppyModule("test")
    module.load_all(quantum)

    guppy.load_pytket("guppy_circ", circ, module)

    @guppy(module)
    def foo(default_reg: array[qubit, 2], 
            extra_reg: array[qubit, 3]) -> array[bool, 3]:
        # Note that the default_reg name is 'q' so it has to come after 'e...' 
        # lexicographically.
        return guppy_circ(extra_reg, default_reg)

    validate(module.compile())


@pytest.mark.skipif(not tket2_installed, reason="Tket2 is not installed")
def test_register_arrays_multiple_measure(validate):
    from pytket import Circuit

    circ = Circuit(2)
    reg1 = circ.add_q_register("extra_reg1", 3)
    reg2 = circ.add_q_register("extra_reg2", 2)
    circ.measure_register(reg1, "extra_bits1")
    circ.measure_register(reg2, "extra_bits2")

    module = GuppyModule("test")
    module.load_all(quantum)

    guppy.load_pytket("guppy_circ", circ, module)

    @guppy(module)
    def foo(default_reg: array[qubit, 2], 
            extra_reg1: array[qubit, 3]) -> tuple[array[bool, 3], array[bool, 2]]:
        extra_reg2 = array(qubit(), qubit())
        result = guppy_circ(extra_reg1, extra_reg2, default_reg)
        # Until we add linearity checks to loaded circuits need to discard owned 
        # arrays.
        discard_array(extra_reg2) 
        return result

    validate(module.compile())


@pytest.mark.skipif(not tket2_installed, reason="Tket2 is not installed")
def test_register_arrays_mixed(validate):
    from pytket import Circuit

    circ = Circuit(2, 1)
    reg = circ.add_q_register("q2", 3)
    circ.measure_register(reg, "c2")
    circ.Measure(0, 0)

    module = GuppyModule("test")
    module.load_all(quantum)

    guppy.load_pytket("guppy_circ", circ, module)

    @guppy(module)
    def foo(q: array[qubit, 2], 
            q2: array[qubit, 3]) -> tuple[array[bool, 1], array[bool, 3]]:
        return guppy_circ(q, q2)

    validate(module.compile())



from collections.abc import Callable

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from tests.util import compile_guppy


def test_basic(validate):
    module = GuppyModule("test")

    @guppy(module)
    def bar(x: int) -> bool:
        return x > 0

    @guppy(module)
    def foo() -> Callable[[int], bool]:
        return bar

    validate(module.compile())


def test_call_1(validate):
    module = GuppyModule("test")

    @guppy(module)
    def bar() -> bool:
        return False

    @guppy(module)
    def foo() -> Callable[[], bool]:
        return bar

    @guppy(module)
    def baz() -> bool:
        return foo()()

    validate(module.compile())


def test_call_2(validate):
    module = GuppyModule("test")

    @guppy(module)
    def bar(x: int) -> Callable[[int], None]:
        return bar(x - 1)

    @guppy(module)
    def foo() -> Callable[[int], Callable[[int], None]]:
        return bar

    @guppy(module)
    def baz(y: int) -> None:
        return foo()(y)(y)

    validate(module.compile())


def test_method(validate):
    module = GuppyModule("module")

    @guppy(module)
    def foo(x: int) -> tuple[int, Callable[[int], int]]:
        f = x.__add__
        return f(1), f

    validate(module.compile())


def test_nested(validate):
    @compile_guppy
    def foo(x: int) -> Callable[[int], bool]:
        def bar(y: int) -> bool:
            return x > y

        return bar

    validate(foo)


def test_nested_capture_struct(validate):
    module = GuppyModule("test")

    @guppy.struct(module)
    class MyStruct:
        x: int

    @guppy(module)
    def foo(s: MyStruct) -> Callable[[int], bool]:
        def bar(y: int) -> bool:
            return s.x > y

        return bar

    validate(module.compile())


def test_curry(validate):
    module = GuppyModule("curry")

    @guppy(module)
    def curry(f: Callable[[int, int], bool]) -> Callable[[int], Callable[[int], bool]]:
        def g(x: int) -> Callable[[int], bool]:
            def h(y: int) -> bool:
                return f(x, y)

            return h

        return g

    @guppy(module)
    def uncurry(
        f: Callable[[int], Callable[[int], bool]],
    ) -> Callable[[int, int], bool]:
        def g(x: int, y: int) -> bool:
            return f(x)(y)

        return g

    @guppy(module)
    def gt(x: int, y: int) -> bool:
        return x > y

    @guppy(module)
    def main(x: int, y: int) -> None:
        curried = curry(gt)
        curried(x)(y)
        uncurried = uncurry(curried)
        uncurried(x, y)
        curry(uncurry(curry(gt)))(y)(x)

    validate(module.compile())


def test_y_combinator(validate):
    module = GuppyModule("fib")

    @guppy(module)
    def fac_(f: Callable[[int], int], n: int) -> int:
        if n == 0:
            return 1
        return n * f(n - 1)

    @guppy(module)
    def Y(f: Callable[[Callable[[int], int], int], int]) -> Callable[[int], int]:
        def y(x: int) -> int:
            return f(Y(f), x)

        return y

    @guppy(module)
    def fac(x: int) -> int:
        return Y(fac_)(x)

    validate(module.compile())



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


module = GuppyModule("test")


@guppy.declare(module)
def foo(x: int @owned) -> qubit: ...


module.compile()



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load_all(quantum)


@guppy.declare(module)
def foo() -> array[qubit, 10]: ...


@guppy(module)
def main() -> qubit:
    return foo()[0]


module.compile()


"""Guppy standard module for quantum operations."""

# mypy: disable-error-code="empty-body, misc, valid-type"

from typing import no_type_check

from hugr import tys as ht

from guppylang.decorator import guppy
from guppylang.std._internal.compiler.quantum import (
    InoutMeasureCompiler,
    RotationCompiler,
)
from guppylang.std._internal.util import quantum_op
from guppylang.std.angles import angle, pi
from guppylang.std.builtins import array, owned
from guppylang.std.option import Option


@guppy.type(ht.Qubit, copyable=False, droppable=False)
class qubit:
    @guppy.hugr_op(quantum_op("QAlloc"))
    @no_type_check
    def __new__() -> "qubit": ...

    @guppy
    @no_type_check
    def measure(self: "qubit" @ owned) -> bool:
        return measure(self)

    @guppy
    @no_type_check
    def project_z(self: "qubit") -> bool:
        return project_z(self)

    @guppy
    @no_type_check
    def discard(self: "qubit" @ owned) -> None:
        discard(self)


@guppy.hugr_op(quantum_op("TryQAlloc"))
@no_type_check
def maybe_qubit() -> Option[qubit]:
    """Try to allocate a qubit, returning `some(qubit)`
    if allocation succeeds or `nothing` if it fails."""


@guppy.hugr_op(quantum_op("H"))
@no_type_check
def h(q: qubit) -> None: ...


@guppy.hugr_op(quantum_op("CZ"))
@no_type_check
def cz(control: qubit, target: qubit) -> None: ...


@guppy.hugr_op(quantum_op("CY"))
@no_type_check
def cy(control: qubit, target: qubit) -> None: ...


@guppy.hugr_op(quantum_op("CX"))
@no_type_check
def cx(control: qubit, target: qubit) -> None: ...


@guppy.hugr_op(quantum_op("T"))
@no_type_check
def t(q: qubit) -> None: ...


@guppy.hugr_op(quantum_op("S"))
@no_type_check
def s(q: qubit) -> None: ...


@guppy.hugr_op(quantum_op("X"))
@no_type_check
def x(q: qubit) -> None: ...


@guppy.hugr_op(quantum_op("Y"))
@no_type_check
def y(q: qubit) -> None: ...


@guppy.hugr_op(quantum_op("Z"))
@no_type_check
def z(q: qubit) -> None: ...


@guppy.hugr_op(quantum_op("Tdg"))
@no_type_check
def tdg(q: qubit) -> None: ...


@guppy.hugr_op(quantum_op("Sdg"))
@no_type_check
def sdg(q: qubit) -> None: ...


@guppy.custom(RotationCompiler("Rz"))
@no_type_check
def rz(q: qubit, angle: angle) -> None: ...


@guppy.custom(RotationCompiler("Rx"))
@no_type_check
def rx(q: qubit, angle: angle) -> None: ...


@guppy.custom(RotationCompiler("Ry"))
@no_type_check
def ry(q: qubit, angle: angle) -> None: ...


@guppy.custom(RotationCompiler("CRz"))
@no_type_check
def crz(control: qubit, target: qubit, angle: angle) -> None: ...


@guppy.hugr_op(quantum_op("Toffoli"))
@no_type_check
def toffoli(control1: qubit, control2: qubit, target: qubit) -> None: ...


@guppy.custom(InoutMeasureCompiler())
@no_type_check
def project_z(q: qubit) -> bool: ...


@guppy.hugr_op(quantum_op("QFree"))
@no_type_check
def discard(q: qubit @ owned) -> None: ...


@guppy.hugr_op(quantum_op("MeasureFree"))
@no_type_check
def measure(q: qubit @ owned) -> bool: ...


@guppy.hugr_op(quantum_op("Reset"))
@no_type_check
def reset(q: qubit) -> None: ...


N = guppy.nat_var("N")


@guppy
@no_type_check
def measure_array(qubits: array[qubit, N] @ owned) -> array[bool, N]:
    """Measure an array of qubits, returning an array of bools."""
    return array(measure(q) for q in qubits)


@guppy
@no_type_check
def discard_array(qubits: array[qubit, N] @ owned) -> None:
    """Discard an array of qubits."""
    for q in qubits:
        discard(q)


# -------NON-PRIMITIVE-------


@guppy
@no_type_check
def ch(control: qubit, target: qubit) -> None:
    # based on https://quantumcomputing.stackexchange.com/a/15737
    ry(target, -pi / 4)
    cz(control, target)
    ry(target, pi / 4)



from hugr import tys

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.option import Option
from guppylang.std.quantum import qubit
from guppylang.std.quantum_functional import h, cx

import guppylang.std.quantum_functional as quantum
from guppylang.tys.ty import NoneType
from tests.util import compile_guppy


def test_basic(validate):
    @compile_guppy
    def test(xs: list[float]) -> list[int]:
        return [int(x) for x in xs]

    validate(test)


def test_basic_linear(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy(module)
    def test(qs: list[qubit] @owned) -> list[qubit]:
        return [h(q) for q in qs]

    validate(module.compile())


def test_guarded(validate):
    @compile_guppy
    def test(xs: list[int]) -> list[int]:
        return [2 * x for x in xs if x > 0 if x < 20]

    validate(test)


def test_multiple(validate):
    @compile_guppy
    def test(xs: list[int], ys: list[int]) -> list[int]:
        return [x + y for x in xs for y in ys if x + y > 42]

    validate(test)


def test_multiple_struct(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy.struct(module)
    class MyStruct:
        qs: list[qubit]

    @guppy(module)
    def test(ss: list[MyStruct] @owned) -> list[qubit]:
        return [h(q) for s in ss for q in s.qs]

    validate(module.compile())


def test_tuple_pat(validate):
    @compile_guppy
    def test(xs: list[tuple[int, int, float]]) -> list[float]:
        return [x + y * z for x, y, z in xs if x - y > z]

    validate(test)


def test_tuple_pat_linear(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy(module)
    def test(qs: list[tuple[int, qubit, qubit]] @owned) -> list[tuple[qubit, qubit]]:
        return [cx(q1, q2) for _, q1, q2 in qs]

    validate(module.compile())


def test_tuple_return(validate):
    @compile_guppy
    def test(xs: list[int], ys: list[float]) -> list[tuple[int, float]]:
        return [(x, y) for x in xs for y in ys]

    validate(test)


def test_dependent(validate):
    module = GuppyModule("test")

    @guppy.declare(module)
    def process(x: float) -> list[int]: ...

    @guppy(module)
    def test(xs: list[float]) -> list[float]:
        return [x * y for x in xs if x > 0 for y in process(x) if y > x]

    validate(module.compile())


def test_capture(validate):
    @compile_guppy
    def test(xs: list[int], y: int) -> list[int]:
        return [x + y for x in xs if x > y]

    validate(test)


def test_capture_struct(validate):
    module = GuppyModule("test")

    @guppy.struct(module)
    class MyStruct:
        x: int
        y: float

    @guppy(module)
    def test(xs: list[int], s: MyStruct) -> list[int]:
        return [x + s.x for x in xs if x > s.y]

    validate(module.compile())


def test_scope(validate):
    @compile_guppy
    def test(xs: list[None]) -> float:
        x = 42.0
        [x for x in xs]
        return x

    validate(test)


def test_nested_left(validate):
    @compile_guppy
    def test(xs: list[int], ys: list[float]) -> list[list[float]]:
        return [[x + y for y in ys] for x in xs]

    validate(test)


def test_nested_right(validate):
    @compile_guppy
    def test(xs: list[int]) -> list[int]:
        return [-x for x in [2 * x for x in xs]]

    validate(test)


def test_nested_linear(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy(module)
    def test(qs: list[qubit] @owned) -> list[qubit]:
        return [h(q) for q in [h(q) for q in qs]]

    validate(module.compile())


def test_classical_list_comp(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy(module)
    def test(xs: list[int]) -> list[int]:
        return [x for x in xs]

    validate(module.compile())


def test_linear_discard(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy.declare(module)
    def discard(q: qubit @owned) -> None: ...

    @guppy(module)
    def test(qs: list[qubit] @owned) -> list[None]:
        return [discard(q) for q in qs]

    validate(module.compile())


def test_linear_discard_struct(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy.struct(module)
    class MyStruct:
        q1: qubit
        q2: qubit

    @guppy.declare(module)
    def discard(q1: qubit @owned, q2: qubit @owned) -> None: ...

    @guppy(module)
    def test(ss: list[MyStruct] @owned) -> list[None]:
        return [discard(s.q1, s.q2) for s in ss]

    validate(module.compile())


def test_linear_consume_in_guard(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy.declare(module)
    def cond(q: qubit @owned) -> bool: ...

    @guppy(module)
    def test(qs: list[tuple[int, qubit]] @owned) -> list[int]:
        return [x for x, q in qs if cond(q)]

    validate(module.compile())


def test_linear_consume_in_iter(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy.declare(module)
    def make_list(q: qubit @owned) -> list[int]: ...

    @guppy(module)
    def test(qs: list[qubit] @owned) -> list[int]:
        return [x for q in qs for x in make_list(q)]

    validate(module.compile())


def test_linear_next_nonlinear_iter(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy.type(NoneType().to_hugr(), module=module)
    class MyIter:
        """An iterator that yields linear values but is not linear itself."""

        @guppy.declare(module)
        def __next__(self: "MyIter") -> Option[tuple[qubit, "MyIter"]]: ...

    @guppy.type(NoneType().to_hugr(), module=module)
    class MyType:
        """Type that produces the iterator above."""

        @guppy.declare(module)
        def __iter__(self: "MyType") -> MyIter: ...

    @guppy(module)
    def test(mt: MyType, xs: list[int]) -> list[tuple[int, qubit]]:
        # We can use `mt` in an inner loop since it's not linear
        return [(x, q) for x in xs for q in mt]

    validate(module.compile())


def test_nonlinear_next_linear_iter(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy.type(
        tys.Opaque(extension="prelude", id="qubit", args=[], bound=tys.TypeBound.Any),
        copyable=False,
        droppable=False,
        module=module,
    )
    class MyIter:
        """A linear iterator that yields non-linear values."""

        @guppy.declare(module)
        def __next__(self: "MyIter" @owned) -> Option[tuple[int, "MyIter"]]: ...

    @guppy.type(NoneType().to_hugr(), module=module)
    class MyType:
        """Type that produces the iterator above."""

        @guppy.declare(module)
        def __iter__(self: "MyType") -> MyIter: ...

    @guppy(module)
    def test(mt: MyType, xs: list[int]) -> list[tuple[int, int]]:
        # We can use `mt` in an outer loop since the target `x` is not linear
        return [(x, x + y) for x in mt for y in xs]

    validate(module.compile())


def test_borrow(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy.declare(module)
    def foo(q: qubit) -> int: ...

    @guppy(module)
    def test(q: qubit, n: int) -> list[int]:
        return [foo(q) for _ in range(n)]

    validate(module.compile())


def test_borrow_nested(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy.declare(module)
    def foo(q: qubit) -> int: ...

    @guppy(module)
    def test(q: qubit, n: int, m: int) -> list[int]:
        return [foo(q) for _ in range(n) for _ in range(m)]

    validate(module.compile())


def test_borrow_guarded(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy.declare(module)
    def foo(q: qubit) -> int: ...

    @guppy(module)
    def test(q: qubit, n: int) -> list[int]:
        return [foo(q) for i in range(n) if i % 2 == 0]

    validate(module.compile())


def test_borrow_twice(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy.declare(module)
    def foo(q: qubit) -> int: ...

    @guppy(module)
    def test(q: qubit, n: int) -> list[int]:
        return [foo(q) + foo(q) for _ in range(n)]

    validate(module.compile())


def test_borrow_in_guard(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy.declare(module)
    def foo(q: qubit) -> int: ...

    @guppy.declare(module)
    def bar(q: qubit) -> bool: ...

    @guppy(module)
    def test(q: qubit, n: int) -> list[int]:
        return [foo(q) for _ in range(n) if bar(q)]

    validate(module.compile())


def test_borrow_in_iter(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy.declare(module)
    def foo(q: qubit) -> int: ...

    @guppy(module)
    def test(q: qubit @ owned) -> tuple[list[int], qubit]:
        return [foo(q) for _ in range(foo(q))], q

    validate(module.compile())


def test_borrow_struct(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy.struct(module)
    class MyStruct:
        q1: qubit
        q2: qubit

    @guppy.declare(module)
    def foo(s: MyStruct) -> int: ...

    @guppy(module)
    def test(s: MyStruct, n: int) -> list[int]:
        return [foo(s) for _ in range(n)]

    validate(module.compile())


def test_borrow_and_consume(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy.declare(module)
    def foo(q: qubit) -> int: ...

    @guppy.declare(module)
    def bar(q: qubit @ owned) -> int: ...

    @guppy(module)
    def test(qs: list[qubit] @ owned) -> list[int]:
        return [foo(q) + bar(q) for q in qs]

    validate(module.compile())





from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit

module = GuppyModule("test")
module.load(qubit)


@guppy.struct(module)
class MyStruct:
    q: qubit


@guppy.declare(module)
def use(q: qubit @owned) -> None: ...


@guppy(module)
def test(s: MyStruct, b: bool) -> None:
    if b:
        use(s.q)


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array

module = GuppyModule("test")


@guppy(module)
def main(x: array[int, bool]) -> None:
    pass


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")

T = guppy.type_var("T", module=module)


@guppy.declare(module)
def foo(x: T, y: T) -> None:
    ...


@guppy(module)
def main(x: bool, y: tuple[bool]) -> None:
    foo(x, y)


module.compile()



from typing import no_type_check

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std import angles, qsystem
from guppylang.std.angles import angle
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit

qsystem_functional = GuppyModule("qsystem_functional")

qsystem_functional.load(qubit, qsystem)
qsystem_functional.load_all(angles)


@guppy(qsystem_functional)
@no_type_check
def phased_x(q: qubit @ owned, angle1: angle, angle2: angle) -> qubit:
    qsystem.phased_x(q, angle1, angle2)
    return q


@guppy(qsystem_functional)
@no_type_check
def zz_phase(q1: qubit @ owned, q2: qubit @ owned, angle: angle) -> tuple[qubit, qubit]:
    qsystem.zz_phase(q1, q2, angle)
    return q1, q2


@guppy(qsystem_functional)
@no_type_check
def measure_and_reset(q: qubit @ owned) -> tuple[qubit, bool]:
    b = qsystem.measure_and_reset(q)
    return q, b


@guppy(qsystem_functional)
@no_type_check
def zz_max(q1: qubit @ owned, q2: qubit @ owned) -> tuple[qubit, qubit]:
    qsystem.zz_max(q1, q2)
    return q1, q2


@guppy(qsystem_functional)
@no_type_check
def rz(q: qubit @ owned, angle: angle) -> qubit:
    qsystem.rz(q, angle)
    return q


@guppy(qsystem_functional)
@no_type_check
def measure(q: qubit @ owned) -> bool:
    result = qsystem.measure(q)
    return result


@guppy(qsystem_functional)
@no_type_check
def qfree(q: qubit @ owned) -> None:
    qsystem.qfree(q)


@guppy(qsystem_functional)
@no_type_check
def reset(q: qubit @ owned) -> qubit:
    qsystem.reset(q)
    return q
